/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                            Find/Replace dialog                                *
*********************************************************************************
*   $Id: FXFindDialog.cpp,v 1.2 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXFindDialog.cpp,v $
* Revision 1.2  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.1.1.1  2000/10/29 14:20:44  tekno
* Initial release
*
*
*
********************************************************************************/

#ifdef __INSURE__
#define __attribute__(x)
#endif

#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXObjectList.h"
#include "FXDict.h"
#include "FXSettings.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXSize.h"
#include "FXPoint.h"
#include "FXRectangle.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXGIFIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXLabel.h"
#include "FXButton.h"
#include "FXMenuButton.h"
#include "FXComposite.h"
#include "FXPacker.h"
#include "FXMatrix.h"
#include "FXShell.h"
#include "FXPopup.h"
#include "FXTopWindow.h"
#include "FXDialogBox.h"
#include "FXScrollbar.h"
#include "FXTextField.h"
#include "FXScrollArea.h"
#include "FXHorizontalFrame.h"
#include "FXSeparator.h"
#include "FXGroupBox.h"
#include "FXCheckButton.h"
#include "FXRadioButton.h"
#include "FXCharset.h" 
#include "FXMessageBox.h"
#include "FXPLText.h"
#include "FXTextSrc.h"
#include "FXFindDialog.h"

// Map
FXDEFMAP(FXFindReplaceDialog) FXFindReplaceDialogMap[]={
  FXMAPFUNCS(SEL_COMMAND,FXFindReplaceDialog::ID_FIND,
                         FXFindReplaceDialog::ID_REPLALL, FXFindReplaceDialog::onCmdAction),
};

// FXFindReplaceDialog implementation
FXIMPLEMENT(FXFindReplaceDialog,FXDialogBox,FXFindReplaceDialogMap,ARRAYNUMBER(FXFindReplaceDialogMap))

void FXFindReplaceDialog::common_init(FXTextSrc * ts) {
  textsrc = ts;
  fd.find = NULL;
  fd.replace = NULL;
  FXColor findbg = FXRGB(191,255,229);
  FXColor replbg = FXRGB(255,192,158);
  
  // Bottom buttons
  buttons2=new FXHorizontalFrame(this,LAYOUT_SIDE_BOTTOM|FRAME_NONE|LAYOUT_FILL_X);

  buttons = new FXMatrix(this,5,LAYOUT_SIDE_BOTTOM|FRAME_NONE|LAYOUT_FILL_X|MATRIX_BY_COLUMNS|PACK_UNIFORM_WIDTH);
  
  // Separator
  //new FXHorizontalSeparator(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|SEPARATOR_GROOVE);
  
  // Contents
  contents=new FXPacker(this,LAYOUT_SIDE_TOP|FRAME_NONE|LAYOUT_FILL_X|LAYOUT_FILL_Y|PACK_UNIFORM_WIDTH);
  fields=new FXMatrix(contents,2,FRAME_NONE|LAYOUT_FILL_X|MATRIX_BY_COLUMNS);
  

  // Buttons: Find, Replace, R+F, F+R, Replace all, Close
  (new FXButton(buttons,"Next",NULL,this,ID_FIND,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(findbg);
  (new FXButton(buttons,"Prev",NULL,this,ID_FIND_PREV,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(findbg);
  (new FXButton(buttons,"First",NULL,this,ID_FIND_FIRST,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(findbg);
  (new FXButton(buttons,"Last",NULL,this,ID_FIND_LAST,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(findbg);
  (new FXButton(buttons,"Find &All",NULL,this,ID_FIND_ALL,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(findbg);
  
  (new FXButton(buttons,"&Replace",NULL,this,ID_REPLACE,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(replbg);
  (new FXButton(buttons,"R,F",NULL,this,ID_REPLFIND,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(replbg);
  (new FXButton(buttons,"F,R",NULL,this,ID_FINDREPL,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(replbg);
  (new FXButton(buttons,"Repl &Sel",NULL,this,ID_REPLALL_SEL,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(replbg);
  (new FXButton(buttons,"Repl All",NULL,this,ID_REPLALL,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X))->setBackColor(replbg);
  
  new FXButton(buttons2,"&Close",NULL,this,ID_CANCEL,FRAME_RAISED|FRAME_THICK|LAYOUT_LEFT);
  move_cursor = new FXCheckButton(buttons2,"Move cursor to last match",NULL,0,ICON_AFTER_TEXT|LAYOUT_RIGHT);

  
  // Text fields and options
  new FXLabel(fields, "Find",NULL,JUSTIFY_RIGHT|LAYOUT_FILL_X);
  find = new FXTextField(fields, 21, this, ID_FIND,TEXTFIELD_NORMAL|TEXTFIELD_ENTER_ONLY|LAYOUT_FILL_COLUMN);
  find->setBackColor(findbg);
  new FXLabel(fields, "Replace",NULL,JUSTIFY_RIGHT|LAYOUT_FILL_X);
  replace = new FXTextField(fields, 21, this, ID_FIND,TEXTFIELD_NORMAL|TEXTFIELD_ENTER_ONLY|LAYOUT_FILL_COLUMN);
  replace->setBackColor(replbg);
  new FXLabel(fields, "Max.",NULL,JUSTIFY_RIGHT|LAYOUT_FILL_X);
  limit = new FXTextField(fields, 6, NULL, 0, TEXTFIELD_NORMAL|TEXTFIELD_INTEGER|LAYOUT_FILL_COLUMN);
  limit->setText("1000");
  /*
  new FXLabel(contents, "Find",NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
  find = new FXTextField(contents, 36, this, ID_FIND,TEXTFIELD_NORMAL|TEXTFIELD_ENTER_ONLY|LAYOUT_SIDE_RIGHT);
  new FXLabel(contents, "Replace",NULL,JUSTIFY_LEFT|LAYOUT_FILL_X);
  replace = new FXTextField(contents, 36, this, ID_FIND,TEXTFIELD_NORMAL|TEXTFIELD_ENTER_ONLY|LAYOUT_SIDE_RIGHT);
  */
  FXHorizontalFrame * hf = new FXHorizontalFrame(contents,LAYOUT_SIDE_BOTTOM|FRAME_NONE|LAYOUT_FILL_X|PACK_UNIFORM_HEIGHT|PACK_UNIFORM_WIDTH);
  
  FXGroupBox * g1 = new FXGroupBox(hf,"Options",
    GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_SIDE_LEFT);

  case_sens = new FXCheckButton(g1,"Case sensitive",NULL,0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  whole_word = new FXCheckButton(g1,"Whole words only",NULL,0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  multi_line = new FXCheckButton(g1,"Multi-line match",NULL,0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);

  FXGroupBox * g2 = new FXGroupBox(hf,"Search+Replace type",
    GROUPBOX_TITLE_LEFT|FRAME_RIDGE|LAYOUT_SIDE_RIGHT); 
  
  literal = new FXRadioButton(g2,"Literal",NULL,0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  c_esc = new FXRadioButton(g2,"C-Escapes",NULL,0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  regex = new FXRadioButton(g2,"Regular expression",NULL,0,ICON_BEFORE_TEXT|LAYOUT_SIDE_TOP);
  literal->setCheck();

  }

// Construct a dialog box
FXFindReplaceDialog::FXFindReplaceDialog(FXWindow * owner, FXTextSrc * ts):
  FXDialogBox(owner,"Find/Replace",DECOR_TITLE|DECOR_BORDER){
  common_init(ts);
  }
  
FXFindReplaceDialog::FXFindReplaceDialog(FXApp * a, FXTextSrc * ts):
  FXDialogBox(a,"Find/Replace",DECOR_TITLE|DECOR_BORDER){
  common_init(ts);
  }


FXFindReplaceDialog::~FXFindReplaceDialog(){
  }

// Set default find text
void FXFindReplaceDialog::setFindText(FXuchar * text, FXint len){
  find->setText(FXString((FXchar *)text, len));
  }

// Set default replace text
void FXFindReplaceDialog::setReplaceText(FXuchar * text, FXint len){
  replace->setText(FXString((FXchar *)text, len));
  }

// Convert string p of length len, specified using ANSI C escape sequences, to an interpreted
// string in ss.  Returns length of converted string if no error, otherwise negative of the
// converted length.  In the case of mis-specification, the result is the most reasonable
// interpretation.  The returned length is guaranteed to be <= len, by the nature of ANSI
// escape sequences.
FXint interpretANSI(FXchar * ss, const FXchar * p, FXint len) {
      FXbool err = FALSE;
      FXint sslen = 0;
      for (FXint i = 0; i < len; i++) {
        if (p[i] != '\\' || i == len-1) {
          ss[sslen++] = p[i];
          if (p[i] == '\\')
            err = TRUE;
          }
        else {
          FXchar c = p[++i];
          switch (c) {
            case '\\': case '?': case '\'': case '"': ss[sslen++] = c; break;
            case 'n': ss[sslen++] = '\n'; break;
            case 't': ss[sslen++] = '\t'; break;
            case 'v': ss[sslen++] = '\v'; break;
            case 'b': ss[sslen++] = '\b'; break;
            case 'r': ss[sslen++] = '\r'; break;
            case 'f': ss[sslen++] = '\f'; break;
            case 'a': ss[sslen++] = '\a'; break;
            case '0': case '1': case '2': case '3':
            case '4': case '5': case '6': case '7':
              ss[sslen] = c - '0';
              if (i < len-1) {
                c = p[++i];
                if (c >= '0' && c <= '7') {
                  ss[sslen] = ss[sslen]<<3 | c-'0';
                  if (i < len-1) {
                    c = p[++i];
                    if (c >= '0' && c <= '7')
                      ss[sslen] = ss[sslen]<<3 | c-'0';
                    else i--;
                    }
                  }
                else i--;
                }
              sslen++;
              break;
            case 'x':
              if (i < len-1 && isxdigit(p[i+1])) {
                ss[sslen] = 0;
                while (i < len-1 && isxdigit(c = p[++i])) {
                  c = toupper(c);
                  ss[sslen] = ss[sslen]<<4 | (isdigit(c) ? c-'0' : c+10-'A');
                  }
                if (i < len-1) i--;
                sslen++;
                break;
                }
              // fall thru
            default:
              // Specification error.
              ss[sslen++] = '\\';
              ss[sslen++] = c;
              err = TRUE;
              break;
            }
          }
        }
      return err ? -sslen : sslen;
      }


FXFindData * FXFindReplaceDialog::getFindData() {
  FXbool ansi = c_esc && c_esc->getCheck();
  
  fd.flen = find->getText().length();
  if (!fd.find)
    FXMALLOC(&fd.find, FXchar, fd.flen);
  else
    FXRESIZE(&fd.find, FXchar, fd.flen);
  
  // Convert find string according to ANSI escapes.
  if (ansi) {
    fd.flen = interpretANSI(fd.find, find->getText().text(), fd.flen);
    if (fd.flen < 0) {
      FXMessageBox::error(this, MBOX_OK,
                  "Error in find string",
                  "Not a valid ANSI C escape sequence");
      return NULL;
      }
    }
  else
    memcpy(fd.find, (FXchar *)find->getText().text(), fd.flen);
  
  fd.fwd = TRUE;	// Set default
  fd.restart = FALSE;	// Set default
  fd.extend = FALSE;	// Set default
  fd.limit = atoi(limit->getText().text());
  if (fd.limit < 1) fd.limit = 1;
  fd.case_sens = case_sens->getCheck();
  fd.whole_word = whole_word->getCheck();
  fd.multi_line = multi_line && multi_line->getCheck();
  fd.move = move_cursor && move_cursor->getCheck();
  if (regex && regex->getCheck()) fd.search_type = FXFindData::FD_REGEX;
  else if (ansi) fd.search_type = FXFindData::FD_C_ESC;
  else fd.search_type = FXFindData::FD_LITERAL;
  return &fd;
  }

FXFindData * FXFindReplaceDialog::getReplaceData() {
  FXbool ansi = c_esc && c_esc->getCheck();
  
  fd.rlen = replace->getText().length();
  if (!fd.replace)
    FXMALLOC(&fd.replace, FXchar, fd.rlen);
  else
    FXRESIZE(&fd.replace, FXchar, fd.rlen);
  if (ansi) {
    fd.rlen = interpretANSI(fd.replace, replace->getText().text(), fd.rlen);
    if (fd.rlen < 0) {
      FXMessageBox::error(this, MBOX_OK,
                  "Error in replace string",
                  "Not a valid ANSI C escape sequence");
      return NULL;
      }
    } 
  else
    memcpy(fd.replace, (FXchar *)replace->getText().text(), fd.rlen);
  if (regex && regex->getCheck()) fd.replace_type = FXFindData::FD_REGEX;
  else if (ansi) fd.replace_type = FXFindData::FD_C_ESC;
  else fd.replace_type = FXFindData::FD_LITERAL;
  return &fd;
  }

long FXFindReplaceDialog::onCmdAction(FXObject* sender,FXSelector sel,void* ptr){
  FXTRACE((50,"%s:onCmdAction %d\n",getClassName(), SELID(sel))); 
  FXFindData * fdp, * rdp;
  FXObject * tgt = getTarget();
  if (!tgt)
    tgt = getOwner();
  fdp = getFindData();
  rdp = getReplaceData();
  switch(SELID(sel)){
    case ID_FIND:
      return fdp ? tgt->handle(this, MKUINT(FXPLText::ID_FINDSTR, SEL_COMMAND), fdp) : 1;
    case ID_FIND_PREV:
      fdp->fwd = FALSE;
      return fdp ? tgt->handle(this, MKUINT(FXPLText::ID_FINDSTR, SEL_COMMAND), fdp) : 1;
    case ID_FIND_FIRST:
      fdp->restart = TRUE;
      return fdp ? tgt->handle(this, MKUINT(FXPLText::ID_FINDSTR, SEL_COMMAND), fdp) : 1;
    case ID_FIND_LAST:
      fdp->fwd = FALSE;
      fdp->restart = TRUE;
      return fdp ? tgt->handle(this, MKUINT(FXPLText::ID_FINDSTR, SEL_COMMAND), fdp) : 1;
    case ID_FIND_ALL:
      return fdp ? tgt->handle(this, MKUINT(FXPLText::ID_FINDALL, SEL_COMMAND), fdp) : 1;
    case ID_REPLACE:
      return rdp ? tgt->handle(this, MKUINT(FXPLText::ID_REPLSTR, SEL_COMMAND), rdp) : 1;
    case ID_REPLFIND:
      return fdp && rdp ? 
               tgt->handle(this, MKUINT(FXPLText::ID_REPLSTR, SEL_COMMAND), rdp) &&
               tgt->handle(this, MKUINT(FXPLText::ID_FINDSTR, SEL_COMMAND), fdp) : 1;
    case ID_FINDREPL:
      return fdp && rdp ?
               tgt->handle(this, MKUINT(FXPLText::ID_FINDSTR, SEL_COMMAND), fdp) &&
               tgt->handle(this, MKUINT(FXPLText::ID_REPLSTR, SEL_COMMAND), rdp) : 1;
    case ID_REPLALL_SEL:
      return rdp ?
               tgt->handle(this, MKUINT(FXPLText::ID_REPLALLSEL, SEL_COMMAND), rdp) : 1;
    case ID_REPLALL:
      return fdp && rdp ?
               tgt->handle(this, MKUINT(FXPLText::ID_REPLALL, SEL_COMMAND), rdp) : 1;
    }
  FXASSERT(0);
  return 0;
  }



