/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                             Page Layout                                       *
*********************************************************************************
*   $Id: FXPageFormat.cpp,v 1.3 2001/12/01 07:48:44 tekno Exp $
*
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: FXPageFormat.cpp,v $
* Revision 1.3  2001/12/01 07:48:44  tekno
* V1.3 initial
*
* Revision 1.2  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.1.1.1  2000/10/29 14:18:10  tekno
* Initial release
*
*
*
********************************************************************************/
#ifdef __INSURE__
#define __attribute__(x)
#define inline
#endif

#include "xincs.h"
#include "fxver.h"
#include "fxdefs.h"
#include "fxkeys.h"
#include "FXStream.h"
#include "FXString.h"
#include "FXObject.h"
#include "FXCharset.h"
#include "FXDict.h"
#include "FXSettings.h"
#include "FXRegistry.h"
#include "FXAccelTable.h"
#include "FXSize.h"
#include "FXPoint.h"
#include "FXRectangle.h"
#include "FXApp.h"
#include "FXId.h"
#include "FXDC.h"
#include "FXDCWindow.h"
#include "FXFont.h"
#include "FXDrawable.h"
#include "FXImage.h"
#include "FXIcon.h"
#include "FXWindow.h"
#include "FXFrame.h"
#include "FXComposite.h"
#include "FXScrollbar.h"
#include "FXPLText.h"
#include "FXShell.h"
#include "FXTextSrc.h"
#include "FXPageFormat.h"
#include "FXTextAnchor.h"
#include "FXTextMap.h"
#include "FXElement.h"

// Required for FXArray
inline void constructElms(FXAnchorInfo*t, unsigned int num) {
  for (unsigned int i = 0; i < num; i++)
    t[i].ta=NULL; 
  }
inline void destructElms(FXAnchorInfo*, unsigned int) { 
  } 


// Object implementation
FXIMPLEMENT(FXPageFormat,FXObject,NULL,0)

FXPageFormat::FXPageFormat() {
  ok = TRUE;
  pltext = NULL;
  textsrc = NULL;
  matrix = NULL;
  attrs = NULL;
  matrow = NULL;
  attrow = NULL;
  cols = 0;
  rows = 0;
  rowvec = NULL;
  toprow = 0;
  flags = 0;
  tabcols = 0;
  defer = FALSE;
  wasmore = FALSE;
  wasanch = FALSE;
  }

FXPageFormat::FXPageFormat(FXPLText * plt, FXTextSrc * ds,
                           FXint opts, FXint r, FXint c) : flags(opts){
  if (!ds) {
    ds = new FXMemTextSrc(NULL, 0);	// FIXME: pass App
    textsrc_private = TRUE;
    ok = ds && ds->isOk();
    }
  else {
    textsrc_private = FALSE;
    ok = ds->isOk();
    }
  pltext = plt;
  textsrc = ds;
  rows = r;
  cols = c;
  matrix = NULL;
  attrs = NULL;
  rowvec = NULL;
  matrow = NULL;
  attrow = NULL;
  if (r * c) (void)makeMatrixAndVector(r, c);
  toprow = 0;
  textsrc->addPageFormat(this);
  tabcols = 8;
  defer = FALSE;
  wasmore = FALSE;
  wasanch = FALSE;
  }

FXint FXPageFormat::makeMatrixAndVector(FXint r, FXint c) {
  if (!(r*c))
    return 1;
  if (matrix) delete[] matrix;
  if (attrs) delete[] attrs;
  if (rowvec) delete[] (rowvec-FXROWINFO_EXTRA);
  if (matrow) delete[] matrow;
  if (attrow) delete[] attrow;
  matrix = new FXchar[r*c];
  attrs = new FXattr[r*c];
  rowvec = (new FXRowInfo[r+1+FXROWINFO_EXTRA*2]) + FXROWINFO_EXTRA;
  matrow = new FXchar[c];
  attrow = new FXattr[c];
  ok = matrix && attrs && rowvec && matrow && attrow;
  return !ok;
  }

inline FXint FXPageFormat::expand(FXint & col, FXchar c) const {
  if (c >= ' ' || c == '\n' || c < 0) {
    col += 1;
    return 1;
    }
  if (c == '\t') {
    FXint n = tabcols - col % tabcols;
    col += n;
    return n;
    }
  col += 2;
  return 2;
  }

FXint FXPageFormat::expand(FXint & col, FXchar c, 
                            FXchar * ptr, FXint len, FXattr * attr) {
  FXint icol = col;
  FXint ret = expand(col, c);
  if (icol < len) {
    if (c >= ' ' || c < 0)
      ptr[icol] = c;
    else if (c == '\n')
      ptr[icol] = c, attr[icol] |= FXPFATTR_SPECIAL;
    else if (c == '\t') {
      ptr[icol] = ret == 1 ? 0x0B : c;
      attr[icol] |= FXPFATTR_SPECIAL;
      for (FXint i = 1; i < ret && icol+i < len; i++) {
        ptr[icol+i] = i+1 < ret ? c : 0x0B;	// Mark last one for arrow head.
        attr[icol+i] |= FXPFATTR_SPECIAL;
        }
      }
    else {
      ptr[icol] = '^';
      attr[icol] |= FXPFATTR_SPECIAL;
      if (icol+1 < len)
        ptr[icol+1] = c + '@', attr[icol+1] |= FXPFATTR_SPECIAL;
      }
    }
  return ret; 
  }

FXint FXPageFormat::nextSplit(FXint & pos, FXint & linect) const {
  // For basic page format, don't do any line wrapping so just return current line length.
  FXint ls = textsrc->getLineOfPos(pos);
  FXint lp = pos;
  pos = textsrc->getPosOfLine(ls+1);
  linect++;
  return pos - lp;
  }
  
FXint FXPageFormat::offs2col(FXint offs, FXint pos) const {
  // Basic page format handles expansion of tabs and control chars.
  const FXchar * ptr;
  FXint len = 0;
  FXint col = 0;
  for (FXint i = 0; i < offs; i++) {
    if (!len) {
      len = textsrc->getBuf(pos, ptr);
      if (!len)
        break;
      }
    expand(col, *ptr);
    ptr++;
    len--;
    }
  return col;
  }
  
FXint FXPageFormat::col2offs(FXint col, FXint pos, FXint len) const {
  // If column is in middle of multi-column char (such as tab), we round to the nearest start-
  // of character.
  const FXchar * ptr;
  FXint p = pos;
  FXint pp = pos;
  FXint blen = 0;
  FXint cc = 0;
  FXint pcc = 0;
  FXint offs = 0;
  while (cc < col && cc < cols && offs < len) {
    pcc = cc;
    while (pos+(FXint)offs >= pp+blen) {
      pp = p;
      blen = textsrc->getBuf(p, ptr);
      if (!blen)
        break;
      }
    expand(cc, ptr[pos+offs-pp]);
    if (cc < cols)
      offs++;
    }
  if (cc > col && cc < cols) {
    if (col - pcc <= cc - col)
      offs--;
    }
  return offs; 
  }

void FXPageFormat::setRowChars(FXint r, FXint pos, FXint len) {
  FXASSERT(r >= toprow && r < toprow+rows);
  FXint i;
  FXint x = rowvec[r-toprow].index;
  FXint cc = 0;
  FXint p;
  const FXchar * ptr;
  FXint blen = 0;
  
  // Copy the previous contents
  memcpy(matrow, matrix+x, cols*sizeof(*matrow));
  memcpy(attrow, attrs+x, cols*sizeof(*attrow));
  
  // Update matrix
  memset(matrix+x, ' ', cols*sizeof(*matrix));
  memset(attrs+x, 0, cols*sizeof(*attrs));
  for (i = 0, p = pos; i < len && cc < cols; i++) {
    if (!blen) {
      blen = textsrc->getBuf(p, ptr);
      if (!blen)
        break;
      }
    expand(cc, *ptr, matrix + x, cols, attrs + x);
    ptr++;
    blen--;
    }

  // This will call markUpdated() for each visible anchor or bracket, but actual graphics
  // rendering is suppressed.
  textsrc->enumerateMarkers(this, pos, pos+len);
  
  // Compare new matrix with old, and set update areas in the widget
  FXushort sc = cols, ec = 0;
  for (i = 0; i < cols; i++)
    if (matrix[x+i] != matrow[i] ||
        attrs[x+i] != attrow[i]) {
      sc = i;
      break;
      }
  if (sc < cols) {
    ec = sc+1;
    for (i = cols-1; i > sc; i--)
    if (matrix[x+i] != matrow[i] ||
        attrs[x+i] != attrow[i]) {
      ec = i+1;
      break;
      }
    }
  if (ec > sc && pltext)
    pltext->setRowDirty(r-toprow, FALSE, sc, ec);
  }

void FXPageFormat::changeCols(FXint newcols) {
  // Basic page format doesn't need to do anything here except set the new cols.
  cols = newcols;
  }

FXint FXPageFormat::changeLines(FXint startline, FXint endline, FXint deltalines,
                                FXint startpos, FXint endpos, FXint deltachars) {
  FXASSERT(startline <= endline);
  FXASSERT(startpos <= endpos);
  return deltalines;
  }
  
void FXPageFormat::refresh() {
  FXint r, rr, tr;	// absolute and relative row numbers; first row considered
  FXint len;
  if (!matrix) return;
  defer = TRUE;

  // First, set up the row info cache
  tr = (FXint)toprow - FXROWINFO_EXTRA;
  FXint pos = tr >= 0 ? getPosOfRow(toprow) : -1;
  wasanch = TRUE;
  newrend.size(0);
  for (r = tr, rr = -FXROWINFO_EXTRA; rr < rows+FXROWINFO_EXTRA; r++, rr++) {
    len = r >= 0 ? getPosOfRow(r+1) - pos : r == -1 ? 1 : 0;
    rowvec[rr].rowlen = len;
    rowvec[rr].pos = pos;
    rowvec[rr].numaux = 0;
    if (r >= 0) {
      rowvec[rr].line = getLineOfRow(r, rowvec[rr].linerow);
      if (rr >= 0 && rr < rows)
        rowvec[rr].index = rr*cols;
      }
    else
      rowvec[rr].line = -1;
    pos += len;
    }
  // Set these fields for 'one past' last cached line (simplifies bookkeeping).
  rowvec[rr].pos = pos;
  rowvec[rr].line = getLineOfRow(r, rowvec[rr].linerow);
  
  // Determine rows which are not the last of a line.
  for (rr = -FXROWINFO_EXTRA; rr < (FXint)rows+FXROWINFO_EXTRA; rr++)
    rowvec[rr].notlast = rowvec[rr].line == rowvec[rr+1].line;

  // Now fill in the matrix information and renderable anchor cache.  This relies
  // on correct row info cache contents, which is why it is done as a separate loop.
  for (r = tr, rr = -FXROWINFO_EXTRA; rr < rows+FXROWINFO_EXTRA; r++, rr++) {
    len = rowvec[rr].rowlen;
    pos = rowvec[rr].pos;
    if (r >= 0) {
      if (rr >= 0 && rr < rows)
        setRowChars(r, pos, len);
      else
        textsrc->enumerateMarkers(this, pos, pos+len);
      }
    }
  
  defer = FALSE;
  updateRenderables();
  }
  
FXint FXPageFormat::getLineOfRow(FXint r, FXint & linerow) const {
  // Basic page format models one line per row.
  linerow = 0;
  return r;
  }
  
FXint FXPageFormat::getRowOfLine(FXint line, FXint & numrows) const {
  numrows = 1;
  return line;
  }

FXint FXPageFormat::numRows() const {
  return textsrc->lines();
  }
  
FXint FXPageFormat::getPosOfRow(FXint r) const {
  return textsrc->getPosOfLine(r);
  }

FXint FXPageFormat::getRowOfPos(FXint pos, FXint & startpos, FXbool atend) const {
  FXint ret = textsrc->getLineOfPos(pos);
  startpos = textsrc->getPosOfLine(ret);
  return ret;
  }

FXint FXPageFormat::getRowOfPos(FXint pos, FXbool atend) const {
  return textsrc->getLineOfPos(pos);
  }

FXint FXPageFormat::rowWidth(FXint r, FXint numchars) const {
  return offs2col(numchars, getPosOfRow(r));
  }

FXint FXPageFormat::rowPos(FXint r, FXint colno, FXbool include_nl) const {
  FXint pos = getPosOfRow(r);
  FXint npos = getPosOfRow(r+1);
  FXint offs = col2offs(colno, pos, (FXint)(npos - pos)); 
  if (!include_nl && offs > 0 &&
     textsrc->getChar(pos+offs-1) == '\n') offs--;	// Adjust if don't want NL
  return pos + (FXint)offs;
  }

void FXPageFormat::preChange(FXint startline, FXint endline, FXint deltalines,
               FXint startpos, FXint endpos, FXint deltachars) {
  }


FXint FXPageFormat::getMatrixCoordsOfPos(FXint pos, FXint & r, FXint & c, FXbool atend) const {
  // This is called frequently, so we make use of the cached information in the rowvec.
  // If the position is outside the visible area, then revert to using getRowOfPos() etc.
  // If the row is out of visible range, the return value is set to -1 or 1, however
  // returned r and c are set properly (i.e. r will be -ve or >= rows) -- this is
  // necessary for correct redraw of cursor etc. since anchors can extend over several
  // rows when rendered.  If the position is further than FXROWINFO_EXTRA rows
  // outside the visible area, then we don't bother calling offs2col() which can be slightly
  // expensive.  This may cause redraw errors for anchors which extend over more than this many
  // display rows from the baseline, but such large anchors would be silly anyway.  To distinguish
  // this case, the return value is set to -2 or 2 which may be interpreted as 'too far outside
  // viewing area to be possibly visible'.
  FXint rowstart;
  
  if (pos < 0)
    pos = 0;
  if (!rowvec || 
      pos < rowvec[-FXROWINFO_EXTRA].pos || 
      pos == rowvec[-FXROWINFO_EXTRA].pos && atend) {
    // Before start, or not yet initialised.
    r = (FXint)(getRowOfPos(pos, rowstart, atend) - toprow);
    }
  else {
    for (r = -FXROWINFO_EXTRA; r < (FXint)rows+FXROWINFO_EXTRA; r++) {
      FXint npos = rowvec[r].pos + (FXint)rowvec[r].rowlen;
      if (pos < npos)
        break;
      if (atend && pos == npos && rowvec[r].notlast) 
        break;
      }
    if (r == (FXint)rows+FXROWINFO_EXTRA) { 
      // After end. 
      r = (FXint)(getRowOfPos(pos, rowstart, atend) - toprow);
      }
    else
      rowstart = rowvec[r].pos;
    }
  if (-FXROWINFO_EXTRA <= r && r < (FXint)rows+FXROWINFO_EXTRA)
    c = offs2col(pos - rowstart, rowstart);
  else {
    c = 0;
    return r < 0 ? -2 : 2;
    }
  return r < 0 ? -1 : r >= (FXint)rows ? 1 : 0;
  }

   
FXint FXPageFormat::resize(FXint r, FXint c) {
  if (!(r*c))
    return 1;
  if (r != rows || c != cols) {
    // Compute new top row.  If columns not changed, use current top row.  Otherwise, make the
    // start of the line visible (this may back the display up a bit if the top line was split).
    FXint newtr = toprow;
    if (c != cols) {
      FXint numrows;
      FXint topline = getFirstLine();
      changeCols(c);
      if (!ok)
        return 1;
      newtr = getRowOfLine(topline, numrows);
      }
    if (makeMatrixAndVector(r, c))
      return 1;
    rows = r;
    toprow = newtr;
    refresh();
    if (pltext) pltext->pagerResize();  // This causes unconditional redraw
    }
  return 0;
  }

FXint FXPageFormat::change(FXint startline, FXint endline, FXint deltalines, 
                           FXint startpos, FXint endpos, FXint deltachars, FXbool more) {
  /* There are several cases to consider:
     1. Change is completely before current display 'window' -- in this case, only the
        top row number might change.
     2. Change is completely after display -- no visible change necessary.
     3. Change within and possibly after display -- retain top row number, but redisplay.
     4. Change is before and within display -- the unchanged portion of the display should
        not move unless necessary.
     Our overall aim is to minimise repositioning of the display, or if it needs to be
     repositioned then do it in the most 'intuitive' way.
  */
  FXbool notified = FALSE;
  FXint dummy;
//#ifndef NDEBUG
#if (0)
  printf("change: startline=%u endline=%u deltalines=%d deltachars=%d more=%d\n",
    startline, endline, deltalines, deltachars, more);
#endif
  FXint dr = changeLines(startline, endline, deltalines, 
                         startpos, endpos, deltachars);	// Delta rows
  if (!ok)
    return 1;
    
  if (more) {
    wasmore = TRUE;
    // Try to keep tabs on where top of display should be.
    if (startline <= getLastLine() && endline < getFirstLine() && dr)
      toprow += dr;
    return 0;
    }
  else if (wasmore) {
    // Be conservative and refresh everything.
    wasmore = FALSE;
    refresh();
    if (pltext) {
      for (FXint i = 0; i < rows; i++)
        pltext->setRowDirty(i, TRUE, cols, 0); // Aux area update req'd
      pltext->pagerUpdate();	// Visible contents change
      }
    return 0;
    }
  
  FXint last_visline = getLastLine();
  if (startline <= last_visline+FXROWINFO_EXTRA) {
    FXint first_visline = getFirstLine();
    if (endline < first_visline-FXROWINFO_EXTRA) {
      toprow += dr;
      for (FXint i = -FXROWINFO_EXTRA; i < (FXint)rows+FXROWINFO_EXTRA; i++) {
        rowvec[i].line += deltalines;
        rowvec[i].pos += deltachars; 
        }
      if (dr && pltext) pltext->pagerLineno();	// Line number change visible only
      notified = TRUE;
      }
    else {
      if (startline < first_visline && endline >= first_visline) {
        if (pltext)
          for (FXint i = 0; i < rows; i++)
            pltext->setRowDirty(i, TRUE, cols, 0); // Aux area update req'd
        toprow += dr;
        }
      refresh();
      if (dr || deltalines) {
        FXint sr = getRowOfLine(startline, dummy) + 1; // Row number of first row to possibly update
        if (pltext && sr >= toprow)
          for (FXint i = sr-toprow; i < rows; i++)
            pltext->setRowDirty(i, TRUE, cols, 0); // Aux area update req'd
        }
      if (pltext) pltext->pagerUpdate();	// Visible contents change
      notified = TRUE;
      }
    }
  if (pltext && dr && !notified)
    pltext->pagerNrows();	// Invisible change (except maybe anchors)
  return 0;
  }

FXint FXPageFormat::removeRenderable(const FXTextAnchor * ta, FXbool temp) {
  //printf("FXPageFormat::removeRenderable %08X temp=%d\n", ta, temp);
  if (ta->isRenderable()) {
    FXint i;
    FXArray<FXAnchorInfo> & arr = temp ? newrend : renderables;
    for (i = 0; i < arr.no(); i++)
      if (arr[i].ta == ta) {
        if (!temp) {
          dirtyAnchor(arr[i], 0);
          if (arr[i].s.aux >= 0 && rowvec) {
            FXint rr = arr[i].s.r;
            for (FXshort k = arr[i].s.aux; k < rowvec[rr].numaux-1; k++)
              rowvec[rr].aux[k] = rowvec[rr].aux[k+1];
            rowvec[rr].numaux--;
            }
          }
        else
          wasanch = TRUE;
        arr.remove(i);
        break;
        }
    }
  return 1;
  }

void FXPageFormat::updateRenderables() {
  if (!wasanch)
    return;
  //printf("FXPageFormat::updateRenderables\n");
  FXint i, j;
  // Clear aux area part of rowvec
  if (rowvec) for (i = -FXROWINFO_EXTRA; i < (FXint)rows+FXROWINFO_EXTRA; i++)
    rowvec[i].numaux = 0;
  for (i = 0; i < renderables.no(); i++)
    renderables[i].visited = FALSE;
  for (i = 0; i < newrend.no(); i++) {
    FXAnchorInfo & nai = newrend[i];
    nai.visited = FALSE;
    for (j = 0; j < renderables.no(); j++)
      if (renderables[j].ta == nai.ta) {
        FXAnchorInfo & oai = renderables[j];
        oai.visited = TRUE;
        nai.visited = TRUE;
        if (rowvec && nai.ta->isAuxRenderable() && rowvec[nai.s.r].numaux < FXROWINFO_MAX_AUX)
          rowvec[nai.s.r].aux[nai.s.aux = rowvec[nai.s.r].numaux++] = i;
        if (oai.s.r != nai.s.r || oai.s.c != nai.s.c) {
          // Changed in main area
          if (oai.s.c >= 0)
            dirtyAnchor(oai, 0);
          if (nai.s.c >= 0)
            dirtyAnchor(nai, 0);
          }
        if (oai.s.aux != nai.s.aux || oai.s.aux >= 0 && oai.s.r != nai.s.r) {
          // Changed in aux area
          if (oai.s.aux >= 0)
            dirtyAnchor(oai, 1);
          if (nai.s.aux >= 0)
            dirtyAnchor(nai, 1);
          }
        break;
        }
    }
  for (i = 0; i < renderables.no(); i++) {
    FXAnchorInfo & oai = renderables[i];
    if (!oai.visited) {
      if (oai.s.c >= 0)
        dirtyAnchor(oai, 0);
      if (oai.s.aux >= 0)
        dirtyAnchor(oai, 1);
      }
    }
  renderables.size(newrend.no());
  for (i = 0; i < newrend.no(); i++) {
    FXAnchorInfo & nai = newrend[i];
    if (!nai.visited) {
      if (rowvec && nai.ta->isAuxRenderable() && rowvec[nai.s.r].numaux < FXROWINFO_MAX_AUX)
          rowvec[nai.s.r].aux[nai.s.aux = rowvec[nai.s.r].numaux++] = i;
      if (nai.s.c >= 0)
        dirtyAnchor(nai, 0);
      if (nai.s.aux >= 0)
        dirtyAnchor(nai, 1);
      }
    renderables[i] = nai;
    //printf("  . %08X: s.r=%d s.c=%d s.aux=%d\n", nai.ta, nai.s.r, nai.s.c, nai.s.aux);
    }
  wasanch = FALSE;
  //printf("-\n");
  }
  
FXint FXPageFormat::addRenderable(const FXTextAnchor * ta, 
     FXshort r1, FXshort c1, FXshort r2, FXshort c2, FXbool temp) {
  //printf("FXPageFormat::addRenderable %08X temp=%d r1=%d c1=%d\n", ta, temp, r1, c1);
  if (!rows) return 0;
  FXASSERT(r1 >= -FXROWINFO_EXTRA && r1 < (FXint)rows+FXROWINFO_EXTRA);
  if (ta->isRenderable()) {
    FXint i;
    FXAnchorInfo ai;
    FXArray<FXAnchorInfo> & arr = temp ? newrend : renderables;
    for (i = 0; i < arr.no(); i++)
      if (arr[i].ta == ta) break;
    ai.ta = (FXTextAnchor *)ta;
    ai.s.r = r1;
    ai.s.c = c1;
    ai.e.r = r2;
    ai.e.c = c2;
    if (!temp && ta->isAuxRenderable() && rowvec && rowvec[r1].numaux < FXROWINFO_MAX_AUX)
      rowvec[r1].aux[ai.s.aux = rowvec[r1].numaux++] = arr.no(); 
    else
      ai.s.aux = -1;
    if (i == arr.no())
      arr.append(ai);
    else
      arr[i] = ai;
    if (!temp)
      dirtyAnchor(ai, 0);
    else
      wasanch = TRUE;
    }
  return 0;
  }

void FXPageFormat::dirtyAnchor(const FXAnchorInfo & ai, FXint where) {
  //printf("FXPageFormat::dirtyAnchor %08X r=%d c=%d where=%d\n", ai.ta, ai.s.r, ai.s.c, where);
  if (pltext) pltext->dirtyAnchor(ai.ta, ai.s.r, ai.s.c, where);
   
  }


FXint FXPageFormat::markUpdated(const FXTextAnchor * ta, FXint sp, FXint ep,
                                FXbool on, FXbool more) { 
  FXint c1, c2, c, clo, chi, x, i;
  FXint r1, r2, r;
  FXint v1;
  //printf("markUpdated bracket %08X sp=%d ep=%d %s%s\n", ta, sp, ep, on ? "ON" : "OFF", more ? " MO" : ""); 
  if (on && !ta->isVisible(pltext)) {
    removeRenderable(ta, TRUE);
    goto show_it;
    }
  v1 = getMatrixCoordsOfPos(sp, r1, c1, !!(ta->getAttrs() & FXTextAnchor::ANCHOR_S_ATEND));
  if (ep > sp) {
    // Is a bracket (non-null range)
    FXint v2 = getMatrixCoordsOfPos(ep-1, r2, c2, !!(ta->getAttrs() & FXTextAnchor::ANCHOR_E_ATEND));
    
    if (!v2) {
      // If last char is the last char of a row, then extend to last column.
      if (getPosOfRow(r2+toprow+1) == ep)
        c2 = cols-1;
      // If last char expands to more than one apparent column, stretch it.
      else {
        expand(c2, textsrc->getChar(ep-1));
        c2--;
        }
      }

    if (v1 < 0 && v2 < -1 ||
        v1 > 1 && v2 > 0) {
      removeRenderable(ta, TRUE);	// Well outside bounds.
      goto show_it;
      }
  
  
    FXattr atr = ta->getAttrs();
    if (!on)
      atr = ~atr;
    
    if (r1 < 0) r1 = 0, c1 = 0;
    if (r2 >= (FXint)rows) r2 = rows-1, c2 = cols-1;
  
    for (r = r1; r <= r2; r++) {
      x = r*cols;
      if (!defer)	// Take a copy for future comparison
        memcpy(attrow, attrs+x, sizeof(*attrow)*cols);
        
      if (r == r1 && r == r2) clo = c1, chi = c2;
      else if (r == r1) clo = c1, chi = cols-1;
      else if (r == r2) clo = 0, chi = c2;
      else clo = 0, chi = cols-1;
      
      if (on)
        for (c = clo; c <= chi; c++)
            attrs[x+c] |= atr;
      else
        for (c = clo; c <= chi; c++)
            attrs[x+c] &= atr;
            
      if (!defer) {
        // Compare new matrix with old, and set update areas in the widget.  We only do this if
        // not defer, since if defer is set then identical processing is already being done
        // by refresh().
        FXushort sc = cols, ec = 0;
        for (i = clo; i <= chi; i++)
          if (attrs[x+i] != attrow[i]) {
            sc = i;
            break;
            }
        if (sc < cols) {
          ec = sc+1;
          for (i = chi; i > clo; i--)
          if (attrs[x+i] != attrow[i]) {
            ec = i+1;
            break;
            }
          }
        if (ec > sc && pltext)
          pltext->setRowDirty(r, FALSE, sc, ec);
        }
      }
    }
  else {
    // Is an anchor (null range).
    if (v1 < -1 || v1 > 1) {
      removeRenderable(ta, TRUE); // Not visible within FXROWINFO_EXTRA rows of top or bottom
      goto show_it;
      }
    }

  on ? addRenderable(ta, r1, c1, r2, c2, TRUE) : removeRenderable(ta, TRUE);

show_it:
  if (!defer && !more) {
    updateRenderables();
    if (pltext) pltext->updateDirty();
    }
  return 0; 
  }

  
FXint FXPageFormat::getCharAttr(FXint r, FXint c, FXchar & ch, FXattr & attr) const {
  if (r < toprow || r >= toprow+rows || c >= cols)
    return 1;
  ch = matrix[(r-toprow)*cols + c];
  attr = attrs[(r-toprow)*cols + c];
  return 0;
  }

FXPLText * FXPageFormat::setPLText(FXPLText * plt) {
  FXPLText * opl = pltext;
  pltext = plt;
  return opl;
  }

void FXPageFormat::setTopRow(FXint r) {
  FXint dr = r - toprow;
  toprow = r;
  refresh();
  if (pltext) {
    pltext->pagerScroll(dr);
    textsrc->handle(this, MKUINT(FXTextSrc::ID_TRACK_TOP, SEL_CHANGED), pltext);
    }
  }

FXint FXPageFormat::getFirstLine() const {
  FXint linerow;
  return getLineOfRow(toprow, linerow);
  }
  
FXint FXPageFormat::getLastLine() const {
  FXint linerow;
  return getLineOfRow(toprow + rows - 1, linerow);
  }
  
FXint FXPageFormat::getFirstPos() const {
  return getPosOfRow(toprow);
  }
  
FXint FXPageFormat::getLastPos() const {
  return getPosOfRow(toprow + rows)-1;
  }

void FXPageFormat::setTabColumns(FXint t) {
  tabcols = t;
  if (matrix) {
    // In general, changing tab columns can require reflow of the entire text, being
    // somewhat equivalent to a change in number of columns.  Thus, we call changeCols()
    // (with the same number of physical columns).  We save the current top row to prevent
    // the display jumping.
    FXint numrows;
    FXint topline = getFirstLine();
    changeCols(cols);
    toprow = getRowOfLine(topline, numrows);
    refresh();
    if (pltext) {
      for (FXint i = 0; i < rows; i++)
        pltext->setRowDirty(i, TRUE, cols, 0); // Aux area update req'd
      pltext->pagerUpdate();	// Visible contents change
      }
    }
  }

  
void FXPageFormat::save(FXStream& /*store*/) const {
  }
  
void FXPageFormat::load(FXStream& /*store*/) {
  }
  
FXPageFormat::~FXPageFormat() {
  if (matrix) delete[] matrix;
  if (attrs) delete[] attrs;
  if (rowvec) delete[] (rowvec-FXROWINFO_EXTRA);
  if (matrow) delete[] matrow;
  if (attrow) delete[] attrow;
  textsrc->removePageFormat(this);
  if (textsrc_private)
    delete textsrc;
  textsrc = (FXTextSrc *)-1;
  }


/*------------------ Pager that implements line wrapping -----------------*/

// Object implementation
FXIMPLEMENT(FXWrapPageFormat,FXPageFormat,NULL,0)

FXWrapPageFormat::FXWrapPageFormat() : FXPageFormat(), delimiters("") {
  map = NULL;
  memset(delimtab, 0, sizeof(delimtab));
  }

FXWrapPageFormat::FXWrapPageFormat(FXPLText * plt, FXTextSrc * ds,
                                   FXint opts, FXint r, FXint c, 
                                   const FXchar * delimset) :
              FXPageFormat(plt, ds, opts, r, c), 
              delimiters(delimset) {
  loadDelimtab();
  map = new FXTextMap(1);
  if (map && map->isOk()) {
    map->updateLeaf(0, 1, 1);	// Set the minimum requirement
    loadMap();			// Initial load from the text source
    }
  else
    ok = FALSE;
  }



FXint FXWrapPageFormat::nextSplit(FXint & pos, FXint & linect) const {
  const FXchar * ptr;	// Current buffer contents
  FXint size;		// Number of chars at above location
  FXint p = pos;	// Next position
  FXint ipos = pos;	// Initial position
  FXint applen = 0;	// Accumulated apparent row length (displayed char positions) -
  			//  this must be <= cols.
  FXint split = -1;	// Position of last candidate line split point, -ve if none determined
  FXint i;
  FXchar c;		// Current char
  FXint q;		// Current position
  
  // NB: we can be called when cols is zero (during initialisation) so in this case we
  // treat this as infinite columns; we will be called again when proper cols has been
  // established.
  FXint vcols = cols;
  if (vcols < 1)
    vcols = (FXint)-1;
  
  do {
    q = p;
    size = textsrc->getBuf(p, ptr); // updates p and ptr
    if (!size) {
      // Reached end of text (initial pos must have been outside buffer).
      pos = p;
      return pos - ipos;
      }
    for (i = 0; i < size; i++, q++) {
      c = ptr[i];
      if (c == '\n')
        break;
      // The following direct implementation changed to use a 256-byte (rather than 256-bit)
      // table because profiler indicated that delimiters.has() was taking a fair chunk of
      // scrolling time.
      //if (delimiters.has(c))
      if (delimtab[(FXuchar)c])
        split = q+1;
      expand(applen, c);
      if (applen >= vcols)
        break;
      }
    } while (i == size);
  if (c == '\n') {
    linect++;
    pos = q+1;
    }
  else if (split > ipos)
    pos = split;
  else {
    // must have exceeded columns with no candidate split
    if (q > ipos)
      pos = q;
    else
      pos = ipos + 1;	// Must put at least one char per line
    }
  return pos - ipos;
  }

void FXWrapPageFormat::changeCols(FXint newcols) {
  cols = newcols;
  loadMap();	// This can set error condition.
  }

void FXWrapPageFormat::preChange(FXint startline, FXint endline, FXint deltalines,
               FXint startpos, FXint endpos, FXint deltachars) {
  /* Compute what we have to change.  This type of page format always starts a new row
     when there is a new line.  A change in any row can potentially affect
     the rest of the rows in the same line.  Thus, we expand the ambit of the change to the
     start of the next line past the ending position.
     
     The information that we need to be save here (and use in changeLines()) consists of the
     position and node numbers of the first row affected, and the start of the next unaffected
     line.
  */
  FXint ilength = map->root_Pn();

  FXASSERT(ilength == textsrc->length());
  FXASSERT(startpos >= 0 && startpos < ilength);
  FXASSERT(endpos >= 0 && endpos < ilength);
  FXASSERT(endpos >= startpos);
  
  // node1 is containing node for line containing start position, node 2 that for end.
  // Ln1 is first row number in node1, Pn1 is length of stretch before replacement start (B).
  lp1 = textsrc->getPosOfLine(startline);
  row1 = getRowOfPos(lp1, node1, Ln1, Pn1);
  lp1 -= Pn1;
  //if (textsrc->getPosOfLine(endline) == endpos)
  //  lp2 = endpos;
  //else
  lp2 = textsrc->getPosOfLine(endline+1); // Start of unaffected tail
  row2 = getRowOfPos(lp2, node2, Ln2, Pn2);
  if (lp2 < ilength)
    lp2 += map->node_Pn(node2) - Pn2;	// Adjust to end of last affected node (+1). 
  }

FXint FXWrapPageFormat::changeLines(FXint startline, FXint endline, FXint deltalines,
                                    FXint startpos, FXint endpos, FXint deltachars) {
  /* This is called by the text source (via change()) after the change has been done.  The
     row map is updated and the nett number of rows inserted is returned.
     
     Nodes [node1..node2] are to be updated.  Corresponding positions in old buffer are
     [lp1..lp2), but lp2 is updated by adding deltachars to correspond to new buffer
     contents.  A new set of nodes replaces the affected range.
  */
  FXint irows = map->root_Ln();
#ifndef NDEBUG
  FXint ilength = map->root_Pn();
#endif
  FXint dr;	// Delta rows (return value)

  FXint u;
  FXint dummy;
  lp2 += deltachars;	// Position one past last char replaced (in new buffer).  
  FXint onodes = (node2 - node1 + 2)>>1;	// Nodes to be replaced

  // Compute how many new nodes
  FXint nrows = 1;
  FXint p = lp1;	// Last position where node starts
  FXint nn = 1;	// # nodes
  FXint len;
  u = lp1;
  for (len = nextSplit(u, dummy); len && u < lp2; len = nextSplit(u, dummy)) {
    if (u > p + 80) p = u, nn++;
    nrows++;
    }
  FXASSERT(u == lp2);
  // Insert or delete nodes in map
  if (nn > onodes) {
    // Note that caller needs to check ok status.
    ok = map->insertLeaves(node1, nn-onodes, 0);
    if (!ok)
      return 0;
    }
  else if (nn < onodes)
    map->deleteLeaves(node1, onodes-nn);
  // Update nodes
  FXint nnd = node1;
  p = lp1;
  FXint ll = 1;
  u = lp1;
  for (len = nextSplit(u, dummy); len && u < lp2; len = nextSplit(u, dummy))
      if (u > p + 80) {
//#ifndef NDEBUG
#if (0)
        printf("  upd leaf %u  Pn=%u Ln=%u\n", nnd, u-p, ll);
#endif
        map->updateLeaf(nnd, u-p, ll); 
        p = u;
        nnd+=2;
        ll = 1;
        }
      else
        ll++; 

  if (u-p) {
    // Update the leaf only if there is non-zero data.  If not, leave the leaf zero.
//#ifndef NDEBUG
#if (0)
    printf("  upd leaf %u  Pn=%u Ln=%u\n", nnd, u-p, ll);
#endif
    map->updateLeaf(nnd, u-p, ll);
    nnd+=2;
  }
  else {
    FXASSERT(0);
    }
  
  FXASSERT(nn == (nnd-node1>>1));
  dr = (FXint)(map->root_Ln() - irows);
    
#ifndef NDEBUG
  if (ilength+deltachars != textsrc->length() ||
      (FXint)map->root_Pn() != textsrc->length()) {
    FXASSERT(0);
    printf("  ilength=%d deltachars=%d Pn=%d ts->length=%d\n",
           ilength,deltachars,map->root_Pn(),textsrc->length());
    }
#endif
  return dr;
  }


void FXWrapPageFormat::loadMap() {
  // Implement as 'replace everything'.
  FXint endpos = map->root_Pn();
  FXint dc = textsrc->length() - endpos;
  node1 = 0; row1 = 0; lp1 = 0;
  node2 = (map->getN() << 1) - 2; row2 = map->root_Ln();
  lp2 = endpos;
  // Note that startline and endline are not used by this class.
  changeLines(0, 0, 0, 0, endpos, dc);	// This can cause error contition
  }

void FXWrapPageFormat::loadDelimtab() {
  // Reload delimtab (cache) based on delimiters.
  for (FXint i = 0; i < 256; i++)
    delimtab[i] = delimiters.has((FXuchar)i);
  }



FXint FXWrapPageFormat::getLineOfRow(FXint r, FXint & linerow) const {
  FXint line = textsrc->getLineOfPos(getPosOfRow(r));
  linerow = r - getRowOfPos(textsrc->getPosOfLine(line));
  return line;
  }
  
FXint FXWrapPageFormat::getRowOfLine(FXint line, FXint & numrows) const {
  FXint row = getRowOfPos(textsrc->getPosOfLine(line)); 
  numrows = getRowOfPos(textsrc->getPosOfLine(line+1)) - row;
  return row;
  }

FXint FXWrapPageFormat::numRows() const {
  return map->root_Ln();
  }

FXint FXWrapPageFormat::getPosOfRow(FXint r) const {
  if (r >= (FXint)map->root_Ln())
    return map->root_Pn();
  FXTextNode search(0, r);
  map->modulo_Ln(search);
  FXint pos = search.Pn;	// Pos of first row of node
  FXint ln = search.Ln;	// Row number within this node
  FXint dummy;
  while (ln) {
    if (!nextSplit(pos, dummy)) break;
    ln--;
    }
  return pos;
  } 

FXint FXWrapPageFormat::getRowOfPos(FXint pos, FXint & startpos, FXbool atend) const {
  if (pos <= 0) {
    startpos = 0;
    return 0;
    }
  if (pos >= (FXint)map->root_Pn())
    pos = map->root_Pn()-1; 
  FXTextNode search((FXint)pos, 0);
  map->modulo_Pn(search);
  FXint spos = pos - (FXint)search.Pn; // Position of 1st row of this node
  FXint sr = search.Ln;	// Row number of 1st row in this node
  FXint dummy;
  startpos = spos;
  while (spos < pos) {
    if (!nextSplit(spos, dummy)) break;
    if (spos <= pos)
      sr++, startpos = spos; 
    }
  if (startpos == pos && atend && textsrc->getChar(pos-1) != '\n') {
    // Need to back up to previous row since this row is not first of a line.
    return getRowOfPos(pos-1, startpos, FALSE); // Recurse is easiest.
    }
  return sr; 
  }

FXint FXWrapPageFormat::getRowOfPos(FXint pos, FXint & node, FXint & Ln, FXint & Pn) const {
  Ln = Pn = node = 0;
  if (pos <= 0) {
    return 0;
    }
  if (pos >= (FXint)map->root_Pn())
    pos = map->root_Pn()-1; 
  FXTextNode search((FXint)pos, 0);
  node = map->modulo_Pn(search);
  Ln = search.Ln;
  Pn = search.Pn;
  FXint spos = pos - (FXint)search.Pn; // Position of 1st row of this node
  FXint sr = search.Ln;	// Row number of 1st row in this node
  FXint dummy;
  while (spos < pos) {
    if (!nextSplit(spos, dummy)) break;
    if (spos <= pos) 
      sr++; 
    }
  return sr;
  }

FXint FXWrapPageFormat::getRowOfPos(FXint pos, FXbool atend) const {
  FXint dummy;
  return getRowOfPos(pos, dummy, atend);
  }

  
void FXWrapPageFormat::setDelimiters(const FXCharset & delim) {
  delimiters = delim;
  loadDelimtab();
  change(0, textsrc->lines()-1, 0, 
         0, textsrc->length()-1, 0);
  }

  

void FXWrapPageFormat::save(FXStream& /*store*/) const {
  }
  
void FXWrapPageFormat::load(FXStream& /*store*/) {
  }
  
FXWrapPageFormat::~FXWrapPageFormat() {
  if (map)
    delete map;
  map = (FXTextMap *)-1;
  }


