/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                        Help window for demo program                           *
*********************************************************************************
*   $Id: PLHelpWindow.cpp,v 1.2 2001/12/01 07:48:45 tekno Exp $
*
*    Copyright (C) 2000, Jeroen van der Zijp
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: PLHelpWindow.cpp,v $
* Revision 1.2  2001/12/01 07:48:45  tekno
* V1.3 initial
*
* Revision 1.1.1.1  2000/10/29 14:19:33  tekno
* Initial release
*
*
*
********************************************************************************/
#ifdef __INSURE__
#define __attribute__(x)
#endif

#include "fx.h"
#include "FXPLText.h"
#include "FXTextSrc.h"
#include "FXTextAnchor.h"
#include "FXPageFormat.h"
#include "FXFindDialog.h"
#include "PLHelpWindow.h"
#include <stdio.h>
#include <stdlib.h>


// Help text
const char help[]=
"\n"
"\n"
"PLText Editor On-line Help\n"
"==========================\n"
"\n\n\n"
"Introduction"
"\n\n"
"PLText is a demonstration of the use of the FXPLText text entry widget.  "
"Although this is a demonstration, you can do serious editing of multiple files with pltext.  "
"Hopefully, you will find pltext is robust enough to be your preferred editor.  "
"pltext has been designed for programming languages, hence the 'PL'.  Several features "
"are provided for the convenience of programmers."
"\n\n"
"You can start pltext with the usual FOX options, plus zero or more files to edit.  If more "
"than one file is specified, a tab for that file is automatically opened."
"\n\n"
"You will notice that pltext does not allow horizontal scrolling.  This is unnecessary, since "
"it always wraps long lines into multiple rows."
"\n\n\n"
"Starting a new document"
"\n\n"
"To start a new document, invoke File->New, or press on the New button on the toolbar.  "
"This creates a new tab in the main area tabbook.  The new tab is then ready for entry of "
"text, or you can select File->Open to open a file for editing.  The current tab can be closed "
"by selecting File->Close, or by pressing F3.  "
"Until the new text is saved, its name is 'untitled-n' where n is a number assigned by pltext"
" to ensure that all new text has a distinct name."
"\n\n\n"
"Opening and Saving Files"
"\n\n"
"To open a file, you can invoke File->Open, or press on the Open button on the toolbar.  "
"This will bring up the standard File Dialog, which allows you to select a file.  "
"To save a file, you can invoke File->Save or File->SaveAs.  "
"The former saves the file back to the same filename, while the latter prompts "
"for an alternative filename."
"\n\n\n"
"The File Browser"
"\n\n"
"An alternative method to open files is the File Browser.  "
"You can display the File Browser by invoking View->FileList.  "
"To open a file using the File Browser, simply click on the file. "
"If there are many files, you may want to limit the number of files displayed "
"by specifying a file pattern in the Filter typein field.  "
"The pattern can be any regular file wildcard expression such as \"*.cpp\".  "
"By default, the File Browser shows all files, i.e. the pattern is \"*\"."
"\n\n\n"
"Recent File Menu"
"\n\n"
"The recent file menu shows files which have been recently visited.  "
"You can quickly get back to a file you've been editing simply by selecting "
"one of these recent files."
"\n\n\n"
"Mouse Selection"
"\n\n"
"You can move the cursor by simply clicking on the desired location with the left mouse button. "
"To highlight some text, press the mouse and drag the mouse while holding the left button. "
"To select text a word at a time, you can double-click and drag; "
"to select text a line at a time, you can triple-click and drag.  "
"Performing a shift-click extends the selection from the last cursor location to the "
"current one."
"\n\n\n"
"Text Drag and Drop"
"\n\n"
"With the middle mouse button, you can perform text drag-and-drop.  "
"Simply press the middle mouse button over a piece of selected text and drag the mouse.  "
"Within the same text window, the drag defaults to a text-movement.  You can change this to "
"a text copy by holding down the control key while you're dragging.  "
"Between one text window and another, the drag defaults to a copy operation.  "
"You can change this to a text movement by holding down the shift key while dragging.  "
"If some text has been selected anywhere on your screen, you can paste this into "
"the text buffer by pressing the middle mouse, moving to the desired insert location, "
"and then releasing the middle mouse."
"\n\n\n"
"Scrolling Text"
"\n\n"
"Using the right mouse button, you can grab the text and scroll it. "
"a right mouse drag is a very convenient way to scroll the text buffer by small amount "
"as the scrolling is one row of text per pixel of vertical mouse movement.  "
"You can of course also use the scroll bars.  Because scrolling becomes awkward when "
"dealing with large amounts of text, you can do a fine scroll or vernier-scroll by "
"holding the shift or control keys while moving the scroll bars."
"\n\n\n"
"The Clipboard"
"\n\n"
"After selecting some text, you can cut or copy this text to the clipboard. "
"A subsequent paste operation will then insert the contents of the clipboard at "
"the current cursor location."
"\n\n\n"
"Find/Replace"
"\n\n"
"Selection of Edit->Find, or pressing Ctl+F, pops up a find/replace dialog.  This allows "
"you to search for plain text, ANSI C-escape sequences, and regular expressions.  If you "
"select some text before pressing Ctl+F, then the selected text will become the default "
"search text.  Alternatively, you can press Ctl+G to rapidly move to the next matching text."
"\n\n"
"Using the find/replace dialog, you can press the Replace buttons to replace the currently "
"selected text with the contents of the replacement string.  Rather than alternately pressing "
"Find and Replace, you can press F,R which does find then replace, or R,F which does replace "
"then find next occurrence.  You can also replace all occurrences using ReplAll.  You can do "
"this with confidence, since the changes can be immediately undone (see next section)."
"\n\n"
"You can search for strings matching a regular expression.  This is performed line-by-line "
"similar to the Unix grep(1) command.  All searches can optionally be performed backwards, without "
"case sensitivity, and look for whole words only."
"\n\n\n"
"Undo/Redo"
"\n\n"
"Each editor window remembers every change made to the text contents.  Changes can be undone in "
"sequence right back to the original file condition.  Press Ctl+Z or F8 to undo the last change, "
"then Ctl-Z again to undo the change before that, and so on.  When undoing many changes, you may "
"not be sure where the next undo will occur, since the display does not move when undoing.  You "
"can press Ctl+Y to move the cursor to the point where the next undo will occur."
"\n\n"
"When you have undone one or more changes, you can immediately undo the effect of the undo (i.e. "
"redo the change).  The keys to perform redo are Shift+Ctl+Z or Shift+F8.  Again, you can redo "
"all changes which were undone by repeatedly pressing Shift+Ctl+Z.  Note that if you perform any "
"editing operation (e.g. inserting a character), then it is not possible to do any more redos."
"\n\n\n"
"Autosave"
"\n\n"
"Using Options->Autosave you can tell pltext to automatically save a backup copy of your edits "
"approximately every 20 seconds.  The backup file is given the same name as the file you are "
"editing, with the extension replaced with '.jic'.  If pltext crashes, then next time you use it "
"to edit that file, it will notice that the autosave file is more recent: you are then given the "
"opportunity of using the backup and continuing where you left off (give or take 20 seconds worth "
"of edits).  Autosave, if selected, is in effect for all files being edited.  Autosave files are "
"placed in the /tmp directory."
"\n\n"
"If pltext cannot write the autosave file, then an error message box will appear.  Autosave will "
"be turned off to prevent recurrence of the same error message.  If this occurs, you should determine "
"why the error occurred and correct the situation.  This is most likely to occur if you do not "
"have write permission to the directory of the file which you are editing."
"\n\n\n"
"Miscellaneous functions"
"\n\n"
"Options->Autoindent: this causes new lines to be indented using the same spaces and tabs "
"as the previous line (the one on which return was pressed).  This is convenient for programmers."
"\n\n"
"Options->SaveSettings: the current selection of colours, options, fonts and other information is "
"written to the FOX registry.  Note that settings are *not* saved automatically when you quit "
"pltext, as was done with the FOX TextEdit demo program.  Some of the settings are global to all "
"pltext documents.  Others are specific to particular file extensions.  pltext maintains registry "
"sections for each unique file extension, including one section for 'no extension' files.  Thus, it "
"is possible to have different background colors etc. for each different file extension.  The settings "
"which are saved for each file extension are:"
"\n\n"
"  . foreground, background, selection foreground, selection background colors.\n"
"  . autoindent, show tabs, show line numbers, tab columns options.\n"
"  . font."
"\n\n"
"The global settings are:"
"\n\n"
"  . position and size of pltext window.\n"
"  . file browser settings.\n"
"  . autosave."
"\n\n"
"View->LineNumbers: causes the left hand side of the text window to display line numbers.  There is "
"also an indicator of the current cursor line (if visible)."
"\n\n"
"View->Split: splits the current text window, giving two views of the same document, one above "
"the other.  This can be useful when alternating edits between two sections of the same document."
"\n\n"
"View->ShowTabs: draws tabs and newlines using distinctive graphics.  These are normally invisible "
"because they are 'whitespace'.  However, it is sometimes important to distinguish these characters "
"when programming."
"\n\n"
"Tabs: the tabs field in the status bar allows you to change the indentation level of tab characters.  "
"The default of 8 is standard, but some text editors (such as Borland's IDE) allow different "
"indentations per tab.  Use of this field will allow such documents to 'look right'."
"\n\n"
"Ctl-E: this key makes pltext look at the current selected text (whether or not in pltext itself).  If "
"this text happens to be a decimal number, then pltext will move the cursor to that line number.  This "
"is useful for relating compiler error messages to locations in the source."
"\n\n\n"
"Keyboard Bindings"
"\n\n"
"The following table lists the keyboard bindings:"
"\n\n"
"Key:\tAction:\n"
"====\t=======\n"
"\n"
"Up\tMove cursor up.\n"
"Shift+Up\tMove cursor up and extend selection.\n"
"Down\tMove cursor down.\n"
"Shift+Down\tMove cursor down and extend selection.\n"
"Left\tMove cursor left.\n"
"Shift+Left\tMove cursor left and extend selection.\n"
"Right\tMove cursor right.\n"
"Shift+Right\tMove cursor right and extend selection.\n"
"Home\tMove cursor to begin of line.\n"
"Shift+Home\tMove cursor to begin of line and extend selection.\n"
"Ctl+Home\tMove cursor to top of text.\n"
"End\tMove cursor to end of line.\n"
"Ctl+End\tMove cursor to bottom of text.\n"
"Shift+End\tMove cursor to end of line and extend selection.\n"
"Page Up\tMove cursor up one page.\n"
"Shift+Page Up\tMove cursor up one page and extend selection.\n"
"Page Down\tMove cursor down one page.\n"
"Shift+Page Down\tMove cursor down one page and extend selection.\n"
"Delete\tDelete character after cursor, or text selection.\n"
"Back Space\tDelete character before cursor, or text selection.\n"
"Ctl+A\tSelect all text.\n"
"Ctl+C or F10\tCopy selected text to clipboard.\n"
"Ctl+V or F11\tPaste text from clipboard.\n"
"Ctl+X or F12\tCut selected text to clipboard.\n"
"Ctl+K\tDelete current line.\n"
"Ctl+H\tScroll current line to top of window.\n"
"Ctl+M\tScroll current line to center of window.\n"
"Ctl+L\tScroll current line to bottom of window.\n"
"Ctl+E\tMove cursor to selected line number.\n"
"Ctl+F\tPopup find/replace dialog.\n"
"Ctl+G or F7\tFind next occurrence.\n"
"Shift+Ctl+G or Shift+F7\tFind previous occurrence.\n"
"Ctl+Z or F8\tUndo last change.\n"
"Shift+Ctl+Z or Shift+F8\tRedo last undo.\n"
"Ctl+Y\tMove cursor to site of next undo.\n"
"Shift+Ctl+Y\tMove cursor to site of next redo.\n"
"Ctl+I\tIndent selected text by 1 space.\n"
"Shift+Ctl+I\tOutdent selected text by 1 space.\n"
"\n\n\n"
"Changing Font"
"\n\n"
"You can change font by invoking the Font Selection Dialog from the Font menu. "
"The Font Dialog displays four list boxes showing the font Family, Weight, Style, "
"and Size of each font.  Note that pltext only allows fixed-width fonts to be selected.  "
"You can narrow down the number of fonts displayed by selecting a specific character "
"set, setwidth, pitch, and whether or not scalable fonts are to be listed only.  "
"The All Fonts checkbutton causes all fonts to be listed. Use this feature if you "
"need to select old-style X11 bitmap fonts.  "
"The Preview window shows a sample of text in the selected font."
"\n\n\n"
"Changing colors"
"\n\n"
"The four color wells on the status line determine the color of the "
"selection background, selection foreground, and normal background and foreground "
"respectively.  Double clicking on a colorwell will bring up the Color Selection "
"Dialog. "
"The Color Selection Dialog supports several ways to select a color: \n"
"\n"
"1. By means of Red, Green, and Blue mixing (additive).\n"
"2. By means of Hue, Saturation, Value mixing.\n"
"3. By means of Cyan, Magenta, and Yellow mixing (subtractive).\n"
"4. By name.\n"
"5. Using one of the predefined color wells.\n"
"\n"
"You can also drag and drop colors, and change the predefined colorwells "
"to hold your own custom color collections."
"\n\n"
;
 


/*******************************************************************************/

  
FXIMPLEMENT(HelpWindow,FXDialogBox,NULL,0)


// Construct
HelpWindow::HelpWindow(FXWindow *owner):
  FXDialogBox(owner,"Help on TextEdit",DECOR_TITLE|DECOR_BORDER|DECOR_RESIZE,0,0,500,600){
  FXVerticalFrame *vertical=new FXVerticalFrame(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  // Editor part
  FXHorizontalFrame *editbox=new FXHorizontalFrame(vertical,LAYOUT_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_SUNKEN|FRAME_THICK,0,0,0,0, 0,0,0,0);
  FXPLText* helptext=new FXPLText(editbox,(FXObject *)NULL,(FXSelector)0,TEXT_READONLY|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  
  // Bottom part
  FXHorizontalFrame *closebox=new FXHorizontalFrame(vertical,LAYOUT_BOTTOM|LAYOUT_FILL_X|PACK_UNIFORM_WIDTH);
  new FXButton(closebox,"&Close",NULL,this,FXDialogBox::ID_ACCEPT,LAYOUT_RIGHT|FRAME_RAISED|FRAME_THICK,0,0,0,0, 20,20,5,5);

  helptext->setText(help,strlen(help));
  helptext->getPageFormat()->setTabColumns(26);

  }


// Clean up
HelpWindow::~HelpWindow(){
  }
