/********************************************************************************
*                                                                               *
*            New  M u l t i - L i n e   T e x t   W i d g e t                   *
*                                                                               *
*                          Demonstration text editor                            *
*********************************************************************************
*   $Id: pltext.cpp,v 1.4 2001/12/01 07:48:45 tekno Exp $
*
*    Copyright (C) 2000, Jeroen van der Zijp
*    Copyright (C) 2000, Stephen J. Hardy
*
*    tekno@users.sourceforge.net
*    16 Powers Place, Latham, A.C.T. 2615, Australia
*
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License as published by
*    the Free Software Foundation; either version 2 of the License, or
*    (at your option) any later version.
*
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*
*    You should have received a copy of the GNU Library General Public License
*    along with this program; if not, write to the Free Software
*    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
* $Log: pltext.cpp,v $
* Revision 1.4  2001/12/01 07:48:45  tekno
* V1.3 initial
*
* Revision 1.3  2000/11/01 11:27:10  tekno
* should compile on Win32, added error handling
*
* Revision 1.2  2000/10/30 12:43:24  tekno
* fixed minor problems noted in initial release
*
* Revision 1.1.1.1  2000/10/29 14:20:22  tekno
* Initial release
*
*
*
********************************************************************************/
#ifdef __INSURE__
#define __attribute__(x)
#endif

#include "fx.h"
#include "fxkeys.h"

#include "FXPLText.h"
#include "FXTextSrc.h"
#include "FXTextAnchor.h"
#include "FXPageFormat.h"
#include "FXFindDialog.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>
#include <signal.h>
#ifndef WIN32
#include <unistd.h>
#endif

#include "PLHelpWindow.h"

#define SAVE_MS 20000	// Autosave timeout in ms.

#define SHARED_FRD	// Define to use shared find/replace dialogues.


/*******************************************************************************/
/* Bit of a kludge to force font dialogue to accept only fixed width fonts */
class FXAPI FXFixedFontDialog : public FXFontDialog {
  FXDECLARE(FXFixedFontDialog)
  FXFixedFontDialog() {}
public:
  FXFixedFontDialog(FXWindow* owner,const FXString& name,FXuint opts=0,FXint x=0,FXint y=0,FXint w=600,FXint h=380);
  };

class FXAPI FXFixedFontSelector : public FXFontSelector {
  FXDECLARE(FXFixedFontSelector)
  FXFixedFontSelector() {}
public:
  FXFixedFontSelector(FXComposite *p,FXObject* tgt=NULL,FXSelector sel=0,FXuint opts=0,FXint x=0,FXint y=0,FXint w=0,FXint h=0);
  long onUpdPitch(FXObject*,FXSelector,void*);
  };

FXIMPLEMENT(FXFixedFontDialog,FXFontDialog,NULL,0)

FXFixedFontDialog::FXFixedFontDialog(FXWindow* owner,const FXString& name,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXFontDialog(owner,name,opts,x,y,w,h){
  delete fontbox;
  fontbox=new FXFixedFontSelector(this,NULL,0,LAYOUT_FILL_X|LAYOUT_FILL_Y);
  fontbox->acceptButton()->setTarget(this);
  fontbox->acceptButton()->setSelector(FXDialogBox::ID_ACCEPT);
  fontbox->cancelButton()->setTarget(this);
  fontbox->cancelButton()->setSelector(FXDialogBox::ID_CANCEL);
  }

FXDEFMAP(FXFixedFontSelector) FXFixedFontSelectorMap[]={
  FXMAPFUNC(SEL_UPDATE,FXFontSelector::ID_PITCH,FXFixedFontSelector::onUpdPitch)
  };


FXIMPLEMENT(FXFixedFontSelector,FXFontSelector,FXFixedFontSelectorMap,ARRAYNUMBER(FXFixedFontSelectorMap))

FXFixedFontSelector::FXFixedFontSelector(FXComposite *p,FXObject* tgt,FXSelector sel,FXuint opts,FXint x,FXint y,FXint w,FXint h):
  FXFontSelector(p, tgt, sel, opts, x, y, w, h) {
  pitch->removeItem(0);
  pitch->removeItem(1);
  pitch->setCurrentItem(0);
  }
long FXFixedFontSelector::onUpdPitch(FXObject*,FXSelector,void*){
  pitch->setCurrentItem(0);
  return 1;
  }

/*******************************************************************************/
class TextWindow;

class Editor : public FXObject {
  FXDECLARE(Editor)
  friend class TextWindow;
  FXbool split;
  FXbool greek;
public:
  // Pass messages to editor first then (if not handled) the tab item
  virtual long onDefault(FXObject*,FXSelector,void*);
protected:
  TextWindow * tw;			// Owner
  FXPLText * editor;			// Primary editor
  FXPLText * editor2;			// Secondary editor for split view
  FXPLText * geditor;			// Primary editor (greek)
  FXPLText * geditor2;			// Secondary editor (greek)
  FXFileTextSrc     *textsrc;		// Text source shared by above
  FXVerticalFrame * vf;
  FXSplitter        *splitter;          // Splitter for split view
  FXSplitter        *gsplitter;         // Splitter for norm/greek
  FXSplitter        *gsplitter2;        // Splitter for norm/greek 2
  FXTabItem         *tab1;		// Tab-book item
  Editor() {}
  Editor(TextWindow * owner, FXTabBook * tabbook, const FXString & file);
  void create();
  void toggleSplit();
  void toggleGreek();
  FXbool isSplit() const { return split; }
  FXbool isGreek() const { return greek; }
  void setBackColor(FXColor c);
  ~Editor();
  };
  

// Mini application object
class TextWindow : public FXMainWindow {
  FXDECLARE(TextWindow)
protected:
  FXMenubar         *menubar;                 // Menu bar
  FXMenuPane        *filemenu;                // File menu
  FXMenuPane        *helpmenu;                // Help menu
  FXToolbar         *toolbar;                 // Tool bar
  FXDelegator       *mdi;		      // Delegator to current editor widget
  FXObjectListOf<Editor> editors;             // Widget cluster for each document open
  FXColorWell       *textwell;                // Text color
  FXColorWell       *backwell;                // Text background color
  FXColorWell       *seltextwell;             // Selected text color
  FXColorWell       *selbackwell;             // Selected text background color
  FXColorWell       *cursorwell;              // Selected cursor color
  FXStatusbar       *statusbar;               // Status bar
  FXVerticalFrame   *treebox;                 // Tree box containing directories/files
  FXDirList         *dirlist;                 // Directory view
  FXSplitter        *splitter;                // Splitter
  FXTextField       *filter;                  // Filter for tree list
  FXRecentFiles      mrufiles;                // Recent files
  FXString           filename;                // Work string variable (actual file is textsrc->getFile())
  FXbool             modified;                // Modified state (last known at GUI update)
  FXTabBook 	* tabbook;
  FXTimer	* savetimer;		      // Autosave
  FXbool	  autosave;			// Whether autosave in effect
#ifdef SHARED_FRD
  FXFindReplaceDialog * frd;
public:
  FXFindReplaceDialog * getFRD() const { return frd; }
#endif
  
protected:
  TextWindow(){}
public:
  long onCmdAbout(FXObject*,FXSelector,void*);
  long onCmdRevert(FXObject*,FXSelector,void*);
  long onUpdRevert(FXObject*,FXSelector,void*);
  long onCmdOpen(FXObject*,FXSelector,void*);
  long onCmdOpenButton(FXObject*,FXSelector,void*);
  long onCmdSave(FXObject*,FXSelector,void*);
  long onUpdSave(FXObject*,FXSelector,void*);
  long onCmdSaveAs(FXObject*,FXSelector,void*);
  long onCmdNew(FXObject*,FXSelector,void*);
  long onCmdNewShell(FXObject*,FXSelector,void*);
  long onCmdFont(FXObject*,FXSelector,void*);
  long onCmdQuit(FXObject*,FXSelector,void*);
  long onCmdPrint(FXObject*,FXSelector,void*);
  long onUpdTitle(FXObject*,FXSelector,void*);
  long onCmdTreeList(FXObject*,FXSelector,void*);
  long onCmdBackColor(FXObject*,FXSelector,void*);
  long onUpdBackColor(FXObject*,FXSelector,void*);
  long onCmdTextColor(FXObject*,FXSelector,void*);
  long onUpdTextColor(FXObject*,FXSelector,void*);
  long onCmdSelBackColor(FXObject*,FXSelector,void*);
  long onUpdSelBackColor(FXObject*,FXSelector,void*);
  long onCmdSelTextColor(FXObject*,FXSelector,void*);
  long onUpdSelTextColor(FXObject*,FXSelector,void*);
  long onCmdCursorColor(FXObject*,FXSelector,void*);
  long onUpdCursorColor(FXObject*,FXSelector,void*);
  long onCmdRecentFile(FXObject*,FXSelector,void*);
  long onCmdAutosave(FXObject*,FXSelector,void*);
  long onUpdAutosave(FXObject*,FXSelector,void*);
  long onCmdSplit(FXObject*,FXSelector,void*);
  long onUpdSplit(FXObject*,FXSelector,void*);
  long onCmdGreek(FXObject*,FXSelector,void*);
  long onUpdGreek(FXObject*,FXSelector,void*);
  long onCmdClose(FXObject*,FXSelector,void*);
  long onCmdCrash(FXObject*,FXSelector,void*);
  long onCmdTab(FXObject*,FXSelector,void*);
  long onTimer(FXObject*,FXSelector,void*);
  long onCmdSaveSettings(FXObject*,FXSelector,void*);
  long onCmdSetAppendable(FXObject*,FXSelector,void*);
  long onCmdSetPrependable(FXObject*,FXSelector,void*);
  long onCmdMode(FXObject*,FXSelector,void*);
  long onUpdMode(FXObject*,FXSelector,void*);

public:
  enum{
    ID_ABOUT=FXMainWindow::ID_LAST,
    ID_OPEN,
    ID_CLOSE,
    ID_OPENBUTTON,
    ID_REVERT,
    ID_SAVE,
    ID_SAVEAS,
    ID_NEW,
    ID_NEW_SHELL,
    ID_TITLE,
    ID_FONT,
    ID_QUIT,
    ID_PRINT,
    ID_TREELIST,
    ID_BACKCOLOR,
    ID_TEXTCOLOR,
    ID_SELBACKCOLOR,
    ID_SELTEXTCOLOR,
    ID_CURSORCOLOR,
    ID_RECENTFILE,
    ID_TOGGLE_AUTOSAVE,
    ID_TOGGLE_SPLIT,
    ID_TOGGLE_GREEK,
    ID_SAVETIMER,
    ID_CRASH,
    ID_TABBOOK,
    ID_SAVE_SETTINGS,
    ID_SET_APPENDABLE,
    ID_SET_PREPENDABLE,
    ID_MODE_AUTO,
    ID_MODE_UNIX,
    ID_MODE_UNIXNL,
    ID_MODE_DOS,
    ID_MODE_DOSNL,
    ID_MODE_BINARY,
    ID_LAST
    };
public:
  TextWindow(FXApp* a);
  virtual void create();
  void start(int& argc,char** argv);
  Editor * currentEditor() const { return (Editor *)mdi->getDelegate(); }
  FXPLText * currentPLText() const;
  FXFileTextSrc * currentTextSrc() const { return currentEditor()->textsrc; }
  FXTabItem * currentTab() const { return currentEditor()->tab1; }
  void newEditor(const FXString & file);
  void readRegistry(Editor * ed, FXString ext);
  void writeRegistry(Editor * ed, FXString ext);
  void writeRegistry();
  };


/*******************************************************************************/
  

// Save as
const unsigned char filesaveas[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf3,0x00,0x00,0x80,0x80,0x80,
  0xb2,0xc0,0xdc,0xff,0xff,0xff,0xff,0xff,0x00,0x00,0x00,0x00,0x80,0x80,0x00,0x00,
  0x00,0xff,0x00,0xff,0xff,0x00,0x00,0x80,0xc0,0xc0,0xc0,0x00,0x80,0x80,0x00,0x00,
  0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,
  0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x04,0x5a,0x10,0xc8,0x19,0xaa,0xbd,0x40,0x0c,
  0x91,0xc1,0xfd,0xc0,0x36,0x10,0x24,0xf1,0x59,0x99,0x58,0x4c,0x2c,0xba,0x19,0x5a,
  0x26,0xcc,0x9c,0x17,0x64,0xc6,0x01,0xcb,0x74,0x77,0xe7,0x07,0x9d,0x0c,0xc6,0x99,
  0x49,0x04,0x40,0x1d,0x12,0xd1,0x29,0x85,0x90,0x41,0xdd,0x81,0x39,0x4b,0x24,0x08,
  0x29,0x68,0x94,0x4a,0xc3,0x8a,0xb4,0x53,0x9e,0x00,0x5b,0x4a,0x86,0x69,0xe3,0x89,
  0x01,0xc1,0x6e,0x8b,0x4d,0x16,0x85,0x02,0x3b,0x41,0xc3,0x31,0x68,0xfb,0x09,0x50,
  0xea,0x93,0x22,0x00,0x3b
  };

// File Save
const unsigned char filesave[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0x00,0xff,0xff,0xff,0xff,0x00,0x80,0x80,0x80,0xff,0xff,0xff,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x42,0x08,0xaa,0xb1,0xfe,0x4c,0x34,0x06,0xa3,0x90,0x20,0xcc,0x1a,0xae,0xc4,0x95,
  0x75,0x6d,0x8f,0xa6,0x79,0x64,0xa6,0x0d,0x44,0x87,0x9a,0x0c,0xc1,0x9a,0xb4,0x3c,
  0x05,0x33,0xad,0xd7,0x43,0x50,0x0c,0xbf,0x20,0x10,0xb8,0x6a,0x0d,0x85,0x42,0x8d,
  0x11,0x79,0xec,0x65,0x7a,0x4c,0x64,0x23,0xb0,0x6c,0x26,0x55,0xd6,0xe6,0x6d,0xc7,
  0x55,0x24,0x00,0x00,0x3b
  };


// File Open
const unsigned char fileopen[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0x00,0xff,0x00,0xff,0xff,0x00,0x80,0x80,0x80,0xff,0xff,0xff,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x45,0x08,0xba,0x1c,0xfe,0x2c,0xae,0x20,0x44,0x90,0x91,0xd6,0x8b,0x95,0xb6,0x1c,
  0x00,0x79,0x95,0x25,0x3e,0x03,0x31,0x3c,0xa5,0x73,0xaa,0xf0,0xda,0xae,0x5e,0xfa,
  0xdc,0xb8,0xa3,0x06,0xc5,0xd0,0xff,0x3e,0x9f,0xc3,0x16,0x04,0x02,0x75,0x81,0xa2,
  0xb2,0x77,0x49,0xf2,0x96,0xc5,0x26,0x21,0x69,0x54,0x36,0xa9,0x50,0xa6,0xe7,0x59,
  0xfd,0x71,0x72,0x60,0x4e,0x02,0x00,0x3b
  };

// Print
const unsigned char print[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0xff,0xff,0xff,0xc0,0xc0,0xc0,0x80,0x80,0x80,0x00,0x80,0x80,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x49,0x08,0x0a,0xd1,0xbd,0x10,0x06,0x41,0x5f,0x94,0x62,0xea,0xbb,0x66,0x75,0x1c,
  0x93,0x09,0x43,0x19,0x84,0x1e,0x55,0x5d,0x8e,0x40,0x04,0x6f,0x66,0x29,0x41,0x91,
  0xaa,0xf5,0x49,0x3b,0x7c,0x7f,0x06,0x25,0x92,0x70,0x30,0x24,0x35,0x0a,0xc8,0xa4,
  0x52,0x79,0xb4,0x39,0x99,0x35,0x5b,0x74,0x49,0x95,0xfa,0xae,0x8f,0x1a,0xa1,0xb0,
  0xed,0x72,0xb9,0xba,0x17,0xd6,0x41,0x00,0x24,0x00,0x00,0x3b
  };

// File new
const unsigned char filenew[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0xff,0xff,0x00,0xff,0xff,0xff,0x80,0x80,0x80,0x00,0x00,0x00,0xc0,0xc0,0xc0,0x00,
  0x00,0x00,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x48,0x08,0x0a,0xb1,0xfe,0xab,0x41,0x28,0x18,0xa8,0xd3,0x85,0x3d,0xfa,0x80,0x5b,
  0x28,0x0e,0x44,0x16,0x16,0x82,0x30,0x08,0xe5,0xa3,0x06,0x68,0xba,0xb2,0xd0,0x10,
  0x08,0x31,0x49,0xec,0x8f,0x9d,0xfe,0x85,0x20,0xe1,0xb3,0xf0,0xfd,0x8e,0x43,0xc7,
  0xec,0xf8,0x4b,0x16,0x99,0x48,0xa2,0x62,0x09,0x75,0x4e,0xa1,0x4d,0x29,0x40,0xb7,
  0xeb,0x76,0xb5,0x5b,0x8f,0xd8,0x03,0x48,0x00,0x00,0x3b
  };

// Cut
const unsigned char cut[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf1,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0x00,0x00,0x80,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,
  0x10,0x00,0x00,0x02,0x29,0x84,0x8f,0xa9,0x10,0xd7,0xeb,0x22,0x34,0x8f,0x4d,0xda,
  0x42,0x9d,0xf9,0x3a,0xed,0x49,0xa1,0x10,0x08,0x21,0x60,0x0a,0xea,0xa4,0xa6,0x86,
  0xa9,0xc0,0x2e,0xba,0xc8,0xf4,0x9d,0xd8,0x28,0x5c,0xf3,0xe7,0x54,0x00,0x00,0x3b
  };
  
// Copy
const unsigned char copy[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf1,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0xff,0xff,0xff,0x00,0x00,0x80,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,
  0x10,0x00,0x00,0x02,0x38,0x84,0x8f,0x69,0xc1,0xad,0x18,0x84,0x14,0xec,0xad,0x49,
  0xa9,0x8d,0x8c,0x8e,0x0f,0x42,0x98,0x30,0x60,0xdf,0xc1,0x35,0xe5,0x54,0x0e,0x97,
  0x49,0x45,0xa0,0x9b,0x32,0xeb,0x48,0xe3,0x71,0x43,0x02,0x8d,0x3a,0x92,0xb8,0x14,
  0xa5,0x9f,0xec,0x71,0x33,0x21,0x67,0x33,0x8b,0xf3,0x99,0x28,0x00,0x00,0x3b
  };
  
  
// Paste
const unsigned char paste[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0xff,0xff,0x00,0xc0,0xc0,0xc0,0x80,0x80,0x00,0x80,0x80,0x80,0x00,
  0x00,0x80,0xff,0xff,0xff,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x4b,0x08,0xba,0xdc,0x1e,0x30,0x38,0x15,0xa1,0x10,0x55,0x52,0x52,0x48,0xc0,0x5f,
  0xd0,0x15,0x9a,0x48,0x0e,0x68,0x2a,0x7a,0x1b,0x99,0x41,0x05,0x49,0x8d,0x5c,0x4d,
  0x97,0x36,0x61,0xec,0xbc,0x51,0xd2,0x05,0xc3,0x61,0x38,0xdc,0xb5,0x6a,0x42,0xe2,
  0x41,0x68,0x00,0xac,0x46,0xc2,0xc8,0x92,0x17,0xc8,0x25,0x95,0xc5,0x27,0x27,0x9a,
  0x59,0x3a,0x2b,0x57,0x6c,0xb3,0xd1,0x2b,0x8f,0x27,0x93,0x04,0x00,0x3b
  };

  
// Penguin
const unsigned char penguin[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x12,0x00,0xf2,0x00,0x00,0xb2,0xc0,0xdc,
  0x80,0x80,0x80,0x00,0x00,0x00,0xc0,0xc0,0xc0,0x10,0x10,0x10,0xff,0xff,0xff,0xe0,
  0xa0,0x08,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x12,0x00,0x00,0x03,
  0x53,0x08,0xba,0x21,0x12,0x2b,0xc6,0xe6,0x9e,0x94,0x62,0x64,0x77,0xa3,0x20,0x4e,
  0x21,0x74,0x8b,0x60,0x9c,0x1a,0xa9,0x98,0xa8,0x45,0xb2,0x85,0x38,0x76,0x4f,0x6c,
  0xbb,0x93,0x60,0xdb,0x0d,0xe4,0xd9,0x83,0x1d,0xe7,0x57,0x18,0x04,0x6f,0xb8,0x4c,
  0xec,0x88,0x9c,0x01,0x0c,0x47,0x66,0xac,0xa2,0x38,0x19,0x76,0x36,0x83,0xc3,0xf0,
  0xb4,0x5e,0x77,0x03,0xaf,0xf8,0x7b,0x13,0x77,0xad,0xd3,0xad,0x75,0x61,0xa5,0x54,
  0x02,0x27,0x45,0x02,0x00,0x3b
  };


// Palette of colors
const unsigned char colorpal[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf2,0x00,0x00,0x80,0x80,0x80,
  0x00,0x00,0x00,0xff,0xff,0x00,0xff,0x00,0xff,0x00,0xff,0x00,0xff,0x00,0x00,0x00,
  0x00,0xff,0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x03,
  0x47,0x08,0x0a,0xd1,0xfb,0x30,0x88,0xe9,0xe0,0x92,0x33,0x07,0xcb,0xf2,0xf8,0xd4,
  0xd3,0x64,0xc2,0x37,0x48,0x55,0x83,0x11,0x04,0x83,0x09,0x9b,0x23,0xb1,0x2d,0x49,
  0x6d,0xdd,0x42,0xc6,0x68,0x67,0xdf,0x23,0x58,0xa7,0x40,0x7c,0x05,0x85,0x12,0x62,
  0xe1,0x65,0xc3,0xa1,0x24,0x86,0x68,0x54,0x73,0xc9,0x48,0x0d,0x54,0xd1,0xf1,0x66,
  0x91,0xa9,0x70,0x11,0x30,0x07,0x90,0x00,0x00,0x3b
  };


// Big time logo
extern const unsigned char bigpenguin[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x30,0x00,0x39,0x00,0xf3,0x00,0x00,0xb2,0xc0,0xdc,
  0x5a,0x52,0x41,0x0e,0x0b,0x0b,0x47,0x3b,0x26,0x71,0x6e,0x67,0xb7,0xb5,0xb0,0xd4,
  0xd2,0xce,0xf9,0xf7,0xf7,0x99,0x8d,0x77,0xa3,0x77,0x1b,0xee,0xba,0x12,0xbe,0xa2,
  0x14,0x8b,0x5e,0x07,0xb3,0x8b,0x27,0xee,0xc7,0x1c,0xd2,0x98,0x0f,0x2c,0x00,0x00,
  0x00,0x00,0x30,0x00,0x39,0x00,0x00,0x04,0xfe,0x10,0xc8,0x49,0xab,0x0d,0x22,0x8b,
  0x61,0xbb,0xff,0x20,0xa6,0x8d,0x1c,0x68,0x9e,0xd3,0x30,0xae,0x19,0x81,0xbe,0x9d,
  0xca,0x66,0x81,0xe8,0xc2,0x38,0x31,0x6b,0x84,0x28,0xe0,0x38,0xd9,0x6e,0x15,0x00,
  0xbe,0x84,0xc3,0x91,0x11,0x95,0x9c,0xdd,0x96,0x9f,0xe1,0x20,0x30,0x40,0x16,0xa1,
  0x51,0x01,0xa1,0xe0,0x1b,0x14,0x0a,0x06,0x9d,0xe6,0x8a,0xed,0x6c,0x0c,0x85,0x83,
  0xcc,0x70,0x40,0x18,0x0c,0xa4,0xb2,0x67,0x5a,0x2d,0x00,0x32,0x07,0x4c,0x40,0xad,
  0x29,0xc9,0x2b,0x74,0x5a,0x77,0x02,0x06,0x6b,0x5d,0x7f,0x16,0x2a,0x05,0x02,0x09,
  0x0a,0x0a,0x0b,0x0b,0x0c,0x1b,0x7c,0x19,0x7e,0x88,0x12,0x2a,0x08,0x0d,0x0a,0x0e,
  0x9d,0x9d,0x0a,0x12,0x2b,0x97,0x14,0x18,0x09,0x0f,0x9c,0x9e,0xa9,0xa2,0xa3,0x12,
  0x54,0x9b,0xa8,0x9e,0x0a,0x0f,0x0f,0x2b,0x4f,0x97,0x18,0x0c,0x8e,0xb1,0x0d,0xb3,
  0xb4,0x1b,0x34,0xac,0x00,0x18,0x08,0x0c,0x0f,0x90,0x09,0xa6,0xc6,0x00,0x2a,0x45,
  0x19,0xc1,0x3a,0x05,0x08,0xc9,0x0f,0x0d,0x0d,0x60,0x68,0x2d,0x95,0xac,0x19,0x6f,
  0xd2,0xc8,0x0d,0x00,0x6c,0x07,0x94,0x3c,0xb7,0x93,0x68,0x5f,0x5f,0xe2,0xe3,0x8b,
  0x71,0x88,0xdc,0xec,0xea,0xe3,0xf3,0xe3,0x48,0x82,0x72,0x2d,0xf4,0xeb,0xf4,0xec,
  0x2c,0x72,0x32,0xfc,0x02,0xf2,0x63,0x61,0x09,0x88,0x88,0x3d,0x02,0x13,0xda,0xdb,
  0xb0,0x44,0xcc,0x8f,0x7d,0x01,0xdf,0x48,0x44,0xe0,0xc4,0x88,0x10,0x30,0x09,0xe7,
  0x49,0x4c,0xb3,0xf0,0x07,0x90,0x3e,0xfe,0x10,0x33,0x1e,0xe8,0x56,0x11,0x86,0x8f,
  0x34,0x22,0x05,0xc2,0x99,0x81,0x63,0x4c,0xca,0x8c,0x77,0xec,0xd9,0x02,0xe1,0x30,
  0xe4,0xcb,0x71,0x2b,0x7f,0x69,0x43,0x21,0x03,0xe1,0x4d,0x81,0x3b,0x5e,0x68,0x40,
  0xf9,0x33,0x60,0xc7,0x99,0x1d,0x1c,0x16,0x55,0x38,0x83,0x4c,0x16,0x3c,0x4b,0x03,
  0x3a,0x74,0xf7,0x41,0xc8,0x80,0xa8,0x52,0x83,0x82,0xe8,0x73,0x15,0x2b,0xbd,0x41,
  0xfe,0xb6,0x56,0xf2,0xe9,0xf5,0xc0,0xd4,0x3e,0x26,0x04,0x30,0x48,0xa0,0xa5,0xec,
  0x3c,0x1f,0x95,0x18,0x82,0x38,0xe5,0x48,0x00,0xd1,0xa8,0x9d,0xe0,0xfe,0x72,0x6a,
  0x81,0x97,0x23,0x59,0x02,0x10,0x60,0x35,0xc0,0xa9,0x01,0x8b,0x04,0x0b,0x3e,0x20,
  0x98,0xf5,0x57,0x01,0xdb,0xae,0x45,0x09,0x2b,0x90,0x34,0x82,0x41,0xa7,0x0e,0xd2,
  0xfc,0x7e,0xe2,0xa4,0x02,0xf2,0x4d,0x00,0x8e,0x12,0x08,0xc9,0xe5,0x88,0x82,0xa6,
  0x5e,0x8d,0x1b,0x2f,0xa8,0xf4,0xd3,0x00,0xe8,0x59,0xc8,0xd6,0xd2,0x55,0x80,0x60,
  0x82,0xdf,0xd4,0x8d,0x3b,0xb1,0xcd,0x70,0x37,0xa1,0xeb,0x53,0xbd,0x18,0x73,0x72,
  0x54,0x1b,0x00,0x02,0x5d,0xa9,0x1d,0xa4,0x76,0xf9,0xb2,0xc0,0x2b,0x47,0xb3,0x1b,
  0x4b,0x58,0x3c,0x3c,0x77,0xee,0xdd,0x64,0x33,0x32,0x78,0x8e,0xbb,0x31,0x02,0xea,
  0xdd,0xab,0x73,0x4a,0xd0,0x5b,0xe5,0x86,0xe8,0x7f,0x95,0xa7,0x06,0x8f,0x5b,0x7d,
  0x63,0x9b,0xf4,0xda,0x05,0xa0,0xeb,0xde,0xba,0x72,0xf6,0xe1,0xab,0x27,0x10,0xe9,
  0xb0,0x58,0xfa,0xff,0x8d,0xf1,0x67,0x82,0x1e,0x72,0xb8,0x89,0x16,0xce,0x1b,0xd1,
  0x10,0x60,0x8f,0x7f,0xf9,0xc9,0x52,0x5b,0x70,0xe1,0xb9,0xf7,0x80,0x68,0x4d,0x64,
  0xc0,0x60,0x7d,0x8e,0x34,0x50,0x9c,0x71,0xdf,0x04,0x27,0x5c,0x63,0x13,0xea,0xb5,
  0xc3,0x76,0xe8,0x51,0x83,0x94,0x04,0x04,0x10,0x80,0xcc,0x8a,0x02,0xce,0xd2,0x40,
  0x02,0x0c,0x58,0x55,0xc5,0x8c,0x55,0xd0,0x90,0x40,0x8b,0xd4,0x30,0x70,0xa2,0x05,
  0x29,0x32,0xb0,0xd6,0x8a,0x3e,0x56,0xb1,0x23,0x05,0x03,0xfc,0x08,0xe4,0x86,0x2f,
  0xf4,0xe0,0x63,0x90,0x7c,0x7d,0x40,0xc5,0x92,0x0c,0x34,0x39,0x41,0x04,0x00,0x3b
  };


// Fonts
const unsigned char fonts[]={
  0x47,0x49,0x46,0x38,0x37,0x61,0x10,0x00,0x10,0x00,0xf0,0x00,0x00,0xb2,0xc0,0xdc,
  0x00,0x00,0x00,0x2c,0x00,0x00,0x00,0x00,0x10,0x00,0x10,0x00,0x00,0x02,0x25,0x84,
  0x8f,0x09,0xc1,0xed,0x8c,0x9a,0x81,0x8b,0xce,0x70,0xac,0x55,0x5b,0x45,0xfc,0x74,
  0xd5,0x32,0x95,0xc9,0x86,0x79,0x63,0xa6,0xae,0xa6,0x87,0xb6,0xb1,0xaa,0x89,0x61,
  0x8b,0xe7,0x6d,0x01,0x00,0x3b
  };
  
  
// Patterns
const FXchar *patterns[]={
  "All Files",          "*",
  "C++ Source Files",   "*.cpp",
  "C Source Files",     "*.c",
  "C++ Header Files",   "*.h",
  "Text Files",         "*.txt",
  "HTML Header Files",  "*.html",
  "Lib Header Files",   "*.lib",
  NULL
  };

// Object implementation
FXIMPLEMENT(Editor,FXObject,NULL,0)

Editor::Editor(TextWindow * owner, FXTabBook * tabbook, const FXString & file) {
  tw = owner;
  
  // Make text source
  textsrc = new FXFileTextSrc(owner->getApp(), 0, file);
  
  tab1=new FXTabItem(tabbook, textsrc->getFile(), NULL);

  vf = new FXVerticalFrame(tabbook,LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_RAISED|FRAME_THICK,0,0,0,0,0,0,0,0);
  // Don't track the splitter (if horizontal), reflowing the text can be too expensive.
  splitter=new FXSplitter(vf,
    LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_VERTICAL|SPLITTER_TRACKING);

  // Make editor window
  gsplitter=new FXSplitter(splitter, LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_REVERSED);
  editor = new FXPLText(gsplitter,textsrc,owner,FXWindow::ID_HOTKEY,
    LAYOUT_FILL_X|LAYOUT_FILL_Y|PLTEXT_SHOW_LINE_NUMBERS);
  
  // Set colours for brackets with ERRORS attribute
  //editor->setColor(FXPLText::PLCOLOR_ERRORS_BG, FXRGB(0,128,0));
  //editor->setColor(FXPLText::PLCOLOR_ERRORS_FG, FXRGB(255,255,255));

#ifdef SHARED_FRD
  editor->setFindReplaceDialog(owner->getFRD());
#endif
  
  editor2 = NULL;
  geditor = NULL;
  geditor2 = NULL;
  split = FALSE;
  greek = FALSE;
  }

Editor::~Editor() {
  delete splitter;	// will delete children
  delete vf;
  delete tab1;
  delete textsrc;	// Delete last
  }

void Editor::create() {
  tab1->create();
  vf->create();
  }


void Editor::toggleSplit() {
  split = !split;
  if (split) {
    FXint ht = splitter->getHeight();
    gsplitter->setHeight(ht/2);
    if (geditor) geditor->setHeight(ht/2);
    // Make editor alternate window
    gsplitter2=new FXSplitter(splitter, LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y);
    editor2 = new FXPLText(gsplitter2, textsrc, tw, FXWindow::ID_HOTKEY,
      LAYOUT_FILL_X|LAYOUT_FILL_Y); 
    gsplitter2->setHeight(ht/2);
    editor2->makeLike(editor);
    gsplitter2->create();
    editor2->create();
#ifdef SHARED_FRD
    editor2->setFindReplaceDialog(editor->getFindReplaceDialog());
#endif
  
    }
  else {
    delete gsplitter2;
    editor2 = geditor2 = NULL;
    gsplitter2 = NULL;
    }
  splitter->recalc();
  }

long Editor::onDefault(FXObject* sender,FXSelector sel,void* ptr){ 
  FXPLText * active = editor2 && editor2->isActive() ? editor2 : editor;
  return active && active->handle(sender,sel,ptr) ||
         tab1 && tab1->handle(sender,sel,ptr);
  }

void Editor::toggleGreek() {
  greek = !greek;
  if (greek) {
    FXint wid = gsplitter->getWidth();
    editor->setWidth(wid * 5/6);
    geditor = new FXPLText(gsplitter, textsrc, tw, FXWindow::ID_HOTKEY,
      LAYOUT_FILL_X|LAYOUT_FILL_Y|PLTEXT_GREEK);
    geditor->setWidth(wid * 1/6);
    geditor->makeLike(editor);
    geditor->setGreek(TRUE, 1, 2, editor);
    geditor->create();
#ifdef SHARED_FRD
    geditor->setFindReplaceDialog(editor->getFindReplaceDialog());
#endif
    }
  else {
    delete geditor;
    geditor = NULL;
    }
  gsplitter->recalc();
  }

void Editor::setBackColor(FXColor c) {
  FXfloat ntsc_lum = FXREDVAL(c) * 0.3 + FXGREENVAL(c) * 0.6 + FXBLUEVAL(c) * 0.1;
  tab1->setBackColor(c);
  tab1->setTextColor(ntsc_lum < 128 ? FXRGB(255,255,255) : FXRGB(0,0,0));
  editor->setBackColor(c);
  if (editor2)
    editor2->setBackColor(c);
  }

/*******************************************************************************/
  
// Map
FXDEFMAP(TextWindow) TextWindowMap[]={
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_ABOUT,        TextWindow::onCmdAbout),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_REVERT,       TextWindow::onCmdRevert),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_REVERT,       TextWindow::onUpdRevert),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_OPEN,         TextWindow::onCmdOpen),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_CLOSE,        TextWindow::onCmdClose),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_OPENBUTTON,   TextWindow::onCmdOpenButton),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_SAVE,         TextWindow::onCmdSave),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_SAVE,         TextWindow::onUpdSave),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_SAVEAS,       TextWindow::onCmdSaveAs),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_NEW,          TextWindow::onCmdNew),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_NEW_SHELL,    TextWindow::onCmdNewShell),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_TITLE,        TextWindow::onUpdTitle),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_FONT,         TextWindow::onCmdFont),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_QUIT,         TextWindow::onCmdQuit),
  FXMAPFUNC(SEL_SIGNAL,  TextWindow::ID_QUIT,         TextWindow::onCmdQuit),
  FXMAPFUNC(SEL_CLOSE,   TextWindow::ID_TITLE,        TextWindow::onCmdQuit),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_PRINT,        TextWindow::onCmdPrint),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_TREELIST,     TextWindow::onCmdTreeList),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_BACKCOLOR,    TextWindow::onCmdBackColor),
  FXMAPFUNC(SEL_CHANGED, TextWindow::ID_BACKCOLOR,    TextWindow::onCmdBackColor),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_BACKCOLOR,    TextWindow::onUpdBackColor),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_SELBACKCOLOR, TextWindow::onCmdSelBackColor),
  FXMAPFUNC(SEL_CHANGED, TextWindow::ID_SELBACKCOLOR, TextWindow::onCmdSelBackColor),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_SELBACKCOLOR, TextWindow::onUpdSelBackColor),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_TEXTCOLOR,    TextWindow::onCmdTextColor),
  FXMAPFUNC(SEL_CHANGED, TextWindow::ID_TEXTCOLOR,    TextWindow::onCmdTextColor),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_TEXTCOLOR,    TextWindow::onUpdTextColor),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_SELTEXTCOLOR, TextWindow::onCmdSelTextColor),
  FXMAPFUNC(SEL_CHANGED, TextWindow::ID_SELTEXTCOLOR, TextWindow::onCmdSelTextColor),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_SELTEXTCOLOR, TextWindow::onUpdSelTextColor),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_CURSORCOLOR,  TextWindow::onCmdCursorColor),
  FXMAPFUNC(SEL_CHANGED, TextWindow::ID_CURSORCOLOR,  TextWindow::onCmdCursorColor),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_CURSORCOLOR,  TextWindow::onUpdCursorColor),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_RECENTFILE,   TextWindow::onCmdRecentFile),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_TOGGLE_AUTOSAVE,  TextWindow::onUpdAutosave),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_TOGGLE_AUTOSAVE,  TextWindow::onCmdAutosave),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_TOGGLE_SPLIT,  TextWindow::onUpdSplit),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_TOGGLE_SPLIT,  TextWindow::onCmdSplit),
  FXMAPFUNC(SEL_UPDATE,  TextWindow::ID_TOGGLE_GREEK,  TextWindow::onUpdGreek),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_TOGGLE_GREEK,  TextWindow::onCmdGreek),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_CRASH,        TextWindow::onCmdCrash),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_TABBOOK,      TextWindow::onCmdTab),
  FXMAPFUNC(SEL_TIMEOUT, TextWindow::ID_SAVETIMER,    TextWindow::onTimer),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_SAVE_SETTINGS,TextWindow::onCmdSaveSettings),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_SET_APPENDABLE,TextWindow::onCmdSetAppendable),
  FXMAPFUNC(SEL_COMMAND, TextWindow::ID_SET_PREPENDABLE,TextWindow::onCmdSetPrependable),
  FXMAPFUNCS(SEL_COMMAND,TextWindow::ID_MODE_AUTO,    TextWindow::ID_MODE_BINARY,
  	TextWindow::onCmdMode),
  FXMAPFUNCS(SEL_UPDATE, TextWindow::ID_MODE_AUTO,    TextWindow::ID_MODE_BINARY,
  	TextWindow::onUpdMode),
  };


// Object implementation
FXIMPLEMENT(TextWindow,FXMainWindow,TextWindowMap,ARRAYNUMBER(TextWindowMap))


// Make some windows
TextWindow::TextWindow(FXApp* a):FXMainWindow(a,"FOX Text Editor: - untitled",NULL,NULL,DECOR_ALL,0,0,850,600,0,0){
  FXIcon *bigp=new FXGIFIcon(getApp(),bigpenguin);
  FXIcon *tinyp=new FXGIFIcon(getApp(),penguin);
  
  // Initialize file name
  filename = "";
  
  setIcon(bigp);
  setMiniIcon(tinyp);

  // Make main window; set myself as the target
  setTarget(this);
  setSelector(ID_TITLE);
  
  // Make color dialog
  FXColorDialog *colordlg=new FXColorDialog(this,"Color Dialog");
  
  // Make menu bar
  menubar=new FXMenubar(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X);
  
  // Make delegator
  mdi = new FXDelegator();
  
  // Tool bar
  FXToolbarShell* dragshell=new FXToolbarShell(this,FRAME_RAISED|FRAME_THICK,100,100);
  toolbar=new FXToolbar(this,dragshell,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|PACK_UNIFORM_WIDTH|PACK_UNIFORM_HEIGHT,0,0,0,0, 4,4,2,2);
  
  // Status bar
  statusbar=new FXStatusbar(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|STATUSBAR_WITH_DRAGCORNER);
  
  // The good old penguin, what would we be without it?
  new FXButton(statusbar,"\tHello, I'm Tux...\nThe symbol for the Linux Operating System.\nAnd all it stands for.",tinyp,NULL,0,LAYOUT_FILL_Y|LAYOUT_RIGHT);
  
  // File menu
  filemenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&File",NULL,filemenu);
    
  // Edit Menu
  FXMenuPane *editmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Edit",NULL,editmenu);
  
  // Options Menu
  FXMenuPane *optionsmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"O&ptions",NULL,optionsmenu);
  
  // View menu
  FXMenuPane *viewmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&View",NULL,viewmenu);
    
  // Help menu
  helpmenu=new FXMenuPane(this);
  new FXMenuTitle(menubar,"&Help",NULL,helpmenu,LAYOUT_RIGHT);

  // Splitter
  splitter=new FXSplitter(this,LAYOUT_SIDE_TOP|LAYOUT_FILL_X|LAYOUT_FILL_Y|SPLITTER_TRACKING);
  
  // Sunken border for tree
  treebox=new FXVerticalFrame(splitter,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
  
  // Make tree
  FXHorizontalFrame* treeframe=new FXHorizontalFrame(treebox,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0, 0,0,0,0);
  dirlist=new FXDirList(treeframe,0,this,ID_TREELIST,DIRLIST_SHOWFILES|TREELIST_BROWSESELECT|TREELIST_SHOWS_LINES|TREELIST_SHOWS_BOXES|LAYOUT_FILL_X|LAYOUT_FILL_Y);
  FXHorizontalFrame* filterframe=new FXHorizontalFrame(treebox,LAYOUT_FILL_X);
  new FXLabel(filterframe,"Filter:");
  filter=new FXTextField(filterframe,10,dirlist,FXDirList::ID_SET_PATTERN,JUSTIFY_RIGHT|FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X);
  
  tabbook=new FXTabBook(splitter,this,ID_TABBOOK,LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_RIGHT);
  // Show column number in status bar
  new FXTextField(statusbar,4,mdi,FXPLText::ID_CURSOR_OFFSET,
  	FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,1,1,0,0);

  // Caption before number  
  new FXLabel(statusbar,"Col:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  
  // Show line number in status bar
  new FXTextField(statusbar,4,mdi,FXPLText::ID_CURSOR_LINE, 
    FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,1,1,0,0);

  // Caption before number  
  new FXLabel(statusbar,"Line:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  
  // Text color
  textwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_TEXTCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  backwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_BACKCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  seltextwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_SELTEXTCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  selbackwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_SELBACKCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  cursorwell=new FXColorWell(statusbar,FXRGB(0,0,0),this,ID_CURSORCOLOR,FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y);

  // Show tab cols in status bar
  new FXTextField(statusbar, 2, mdi, FXPLText::ID_TAB_COLUMNS, 
    FRAME_SUNKEN|FRAME_THICK|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,1,1,0,0);

  // Caption before number  
  new FXLabel(statusbar,"Tabs:", NULL, LAYOUT_RIGHT|LAYOUT_CENTER_Y);
  
  
  textwell->setTipText("Text");
  backwell->setTipText("Background");
  seltextwell->setTipText("Selection");
  selbackwell->setTipText("Selection Background");
  cursorwell->setTipText("Cursor");
  
  textwell->setHelpText("Change text color.");
  backwell->setHelpText("Change background color.");
  seltextwell->setHelpText("Change selection color.");
  selbackwell->setHelpText("Change selection background color.");
  cursorwell->setHelpText("Change cursor color.");
  
  // Make some icons
  FXIcon *filenewicon=new FXGIFIcon(getApp(),filenew);
  FXIcon *fileopenicon=new FXGIFIcon(getApp(),fileopen);
  FXIcon *filesaveicon=new FXGIFIcon(getApp(),filesave);
  FXIcon *filesaveasicon=new FXGIFIcon(getApp(),filesaveas,0,IMAGE_ALPHAGUESS);
  FXIcon *printicon=new FXGIFIcon(getApp(),print);
  FXIcon *cuticon=new FXGIFIcon(getApp(),cut);
  FXIcon *copyicon=new FXGIFIcon(getApp(),copy);
  FXIcon *pasteicon=new FXGIFIcon(getApp(),paste);
  
  // Toobar buttons: File manipulation
  new FXButton(toolbar,"&New\tNew\tCreate new document.",filenewicon,this,ID_NEW,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"&Open\tOpen\tOpen document file.",fileopenicon,this,ID_OPENBUTTON,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"&Save\tSave\tSave document.",filesaveicon,this,ID_SAVE,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"Save &As\tSave As\tSave document to another file.",filesaveasicon,this,ID_SAVEAS,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  
  // Toobar buttons: Print
  new FXFrame(toolbar,LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT,0,0,20,20);
  new FXButton(toolbar,"Print\tPrint\tPrint document.",printicon,this,ID_PRINT,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  
  // Toobar buttons: Editing
  new FXFrame(toolbar,LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT,0,0,20,20);
  FXButton * cut_btn = new FXButton(toolbar,"Cut\tCut",cuticon,mdi,FXPLText::ID_CUT_SEL,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  FXButton * copy_btn = new FXButton(toolbar,"Copy\tCopy",copyicon,mdi,FXPLText::ID_COPY_SEL,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  FXButton * paste_btn = new FXButton(toolbar,"Paste\tPaste",pasteicon,mdi,FXPLText::ID_PASTE_SEL,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);

  copy_btn->addHotKey(fxparseaccel("F10"));
  paste_btn->addHotKey(fxparseaccel("F11"));
  cut_btn->addHotKey(fxparseaccel("F12"));
  
  // Color
  new FXFrame(toolbar,LAYOUT_TOP|LAYOUT_LEFT|LAYOUT_FIX_WIDTH|LAYOUT_FIX_HEIGHT,0,0,20,20);
  new FXButton(toolbar,"&Colors\tColors\tDisplay color dialog.",new FXGIFIcon(getApp(),colorpal),colordlg,FXWindow::ID_SHOW,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  new FXButton(toolbar,"Fon&ts\tFonts\tDisplay font dialog.",new FXGIFIcon(getApp(),fonts),this,ID_FONT,ICON_ABOVE_TEXT|BUTTON_TOOLBAR|FRAME_RAISED|LAYOUT_TOP|LAYOUT_LEFT);
  
  // File Menu entries 
  new FXMenuCommand(filemenu,"&Open...\tCtl-O\tOpen document file.",fileopenicon,this,ID_OPENBUTTON);
  new FXMenuCommand(filemenu,"&Close\tF3\tClose current document.",NULL,this,ID_CLOSE); 
  //new FXMenuCommand(filemenu,"&Revert...\t\tRevert to saved.",NULL,this,ID_REVERT);
  new FXMenuCommand(filemenu,"&New\tCtl-N\tCreate new document.",filenewicon,this,ID_NEW); 
  new FXMenuCommand(filemenu,"New Shell\t\tCreate new command line interface.",NULL,this,ID_NEW_SHELL); 
  new FXMenuCommand(filemenu,"&Save\tCtl-S\tSave changes to file.",filesaveicon,this,ID_SAVE);
  new FXMenuCommand(filemenu,"Save &As...\t\tSave document to another file.",filesaveasicon,this,ID_SAVEAS);
  new FXMenuCommand(filemenu,"&Print...\tCtl-P\tPrint document.",printicon,this,ID_PRINT);
  new FXMenuCommand(filemenu,"&Editable\t\tDocument editable.",NULL,mdi,FXPLText::ID_TOGGLE_EDITABLE);
  new FXMenuCommand(filemenu,"Append\t\tDocument appendable only.",NULL,this,ID_SET_APPENDABLE);
  new FXMenuCommand(filemenu,"Prepend\t\tDocument prependable only.",NULL,this,ID_SET_PREPENDABLE);
  new FXMenuCommand(filemenu,"&Iconify...\t\tIconify editor.",NULL,this,FXTopWindow::ID_ICONIFY);
  //new FXMenuCommand(filemenu,"Crash!\t\tDeliberately crash for testing.",NULL,this,ID_CRASH); 
  FXMenuPane * modemenu = new FXMenuPane(this);
    new FXMenuCommand(modemenu,"Auto\t\tDefault line ends.",NULL,this,ID_MODE_AUTO);
    new FXMenuCommand(modemenu,"Unix (\\n)\t\tUnix-style line ends.",NULL,this,ID_MODE_UNIX);
    new FXMenuCommand(modemenu,"Unix + NL\t\tUnix-style line ends; force NL at EOF.",NULL,this,ID_MODE_UNIXNL);
    new FXMenuCommand(modemenu,"DOS (\\r\\n)\t\tDOS-style line ends.",NULL,this,ID_MODE_DOS);
    new FXMenuCommand(modemenu,"DOS + NL\t\tDOS-style line ends; force CRLF at EOF.",NULL,this,ID_MODE_DOSNL);
    new FXMenuCommand(modemenu,"Binary (R/O)\t\tNo line ends, read-only",NULL,this,ID_MODE_BINARY);
  new FXMenuCascade(filemenu,"&Mode\t\tSet modes for saving file.",NULL,modemenu);
  
  // Recent file menu; this automatically hides if there are no files
  FXMenuSeparator* sep1=new FXMenuSeparator(filemenu);
  sep1->setTarget(&mrufiles);
  sep1->setSelector(FXRecentFiles::ID_ANYFILES);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_1);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_2);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_3);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_4);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_5);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_6);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_7);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_8);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_9);
  new FXMenuCommand(filemenu,NULL,NULL,&mrufiles,FXRecentFiles::ID_FILE_10);
  new FXMenuCommand(filemenu,"Clear Recent &Files",NULL,&mrufiles,FXRecentFiles::ID_CLEAR);
  FXMenuSeparator* sep2=new FXMenuSeparator(filemenu);
  sep2->setTarget(&mrufiles);
  sep2->setSelector(FXRecentFiles::ID_ANYFILES);
  new FXMenuCommand(filemenu,"&Quit\tCtl-Q",NULL,this,ID_QUIT);
  
  // Edit Menu entries
  new FXMenuCommand(editmenu,"&Copy\tCtl-C\tCopy selection to clipboard.",copyicon,mdi,FXPLText::ID_COPY_SEL);
  new FXMenuCommand(editmenu,"Cu&t\tCtl-X\tCut selection to clipboard.",cuticon,mdi,FXPLText::ID_CUT_SEL);
  new FXMenuCommand(editmenu,"&Paste\tCtl-V\tPaste from clipboard.",pasteicon,mdi,FXPLText::ID_PASTE_SEL);
  new FXMenuCommand(editmenu,"&Delete\t\tDelete selection.",NULL,mdi,FXPLText::ID_DELETE_SEL);
  new FXMenuCommand(editmenu,"&Find...\tCtl-F\tFind/Replace.",NULL,mdi,FXPLText::ID_FIND);
  new FXMenuCommand(editmenu,"Find &Next\tCtl-G\tFind next matching selection.",NULL,mdi,FXPLText::ID_FINDNEXT);
  new FXMenuCommand(editmenu,"Find P&rev\tShift-Ctl-G\tFind previous matching selection.",NULL,mdi,FXPLText::ID_FINDPREV);
  new FXMenuCommand(editmenu,"&Undo\tCtl-Z\tUndo last change.",NULL,mdi,FXPLText::ID_ROLLBACK);
  new FXMenuCommand(editmenu,"&Redo\tShift-Ctl-Z\tRedo last undo.",NULL,mdi,FXPLText::ID_ROLLFORWARD);
  
  // Options menu
  new FXMenuCommand(optionsmenu,"F&ont...\t\tChange text font.",NULL,this,ID_FONT);
  new FXMenuCommand(optionsmenu,"&Colors...\t\tChange text colors.",NULL,colordlg,FXWindow::ID_SHOW);
  new FXMenuCommand(optionsmenu,"Overstrike\t\tToggle overstrike mode.",NULL,mdi,FXPLText::ID_TOGGLE_OVERSTRIKE);
  new FXMenuCommand(optionsmenu,"&Flash\t\tFlash cursor.",NULL,mdi,FXPLText::ID_TOGGLE_FLASH);
  new FXMenuCommand(optionsmenu,"Auto&indent\t\tToggle automatic line indentation.",NULL,mdi,FXPLText::ID_TOGGLE_AUTOINDENT);
  new FXMenuCommand(optionsmenu,"Autosave\t\tToggle automatic file save.",NULL,this,ID_TOGGLE_AUTOSAVE);
  new FXMenuCommand(optionsmenu,"&Save Settings\t\tSave settings now.",NULL,this,ID_SAVE_SETTINGS);
  
  // View Menu entries
  new FXMenuCommand(viewmenu,"&File list\t\tDisplay file list.",NULL,treebox,FXWindow::ID_TOGGLESHOWN);
  new FXMenuCommand(viewmenu,"Show &hidden files\t\tShow hidden files and directories.",NULL,dirlist,FXDirList::ID_TOGGLE_HIDDEN);
  new FXMenuCommand(viewmenu,"Tool&bar\t\tDisplay toolbar.",NULL,toolbar,FXWindow::ID_TOGGLESHOWN);
  new FXMenuCommand(viewmenu,"&Status line\t\tDisplay status line.",NULL,statusbar,FXWindow::ID_TOGGLESHOWN);
  new FXMenuCommand(viewmenu,"&Line numbers\tCtl-N\tDisplay line numbers.",NULL,mdi,FXPLText::ID_TOGGLE_SHOWAUX);
  new FXMenuCommand(viewmenu,"&Split\t\tShow two views of same text.",NULL,this,ID_TOGGLE_SPLIT);
  new FXMenuCommand(viewmenu,"Show &tabs\tCtl-T\tToggle rendering of tabs and newlines.",NULL,mdi,FXPLText::ID_TOGGLE_SHOWTABS);
  new FXMenuCommand(viewmenu,"&Greek text\t\tShow text map.",NULL,this,ID_TOGGLE_GREEK);
  
  HelpWindow * helpwindow = new HelpWindow(this);
  // Help Menu entries 
  new FXMenuCommand(helpmenu,"&Help...\t\tDisplay help information.",NULL,helpwindow,ID_SHOW,0);
  new FXMenuSeparator(helpmenu);
  new FXMenuCommand(helpmenu,"&About FOX...",NULL,this,ID_ABOUT,0);
  
  // Make a tool tip
  new FXTooltip(getApp(),0); 
  
  // Recent files
  mrufiles.setTarget(this);
  mrufiles.setSelector(ID_RECENTFILE);

#ifdef SHARED_FRD
  frd = new FXFindReplaceDialog(getApp(), NULL);
  frd->setTarget(mdi);
#endif
  
  // Create first editor cluster
  editors.append(new Editor(this, tabbook, FXString("")));
  mdi->setDelegate(editors[0]);

  modified = TRUE;
  
  }

void TextWindow::newEditor(const FXString & file) {
  Editor * ed;
  editors.append(ed = new Editor(this, tabbook, file));
  ed->create();
  ed->editor->makeLike(currentPLText()); 
  ed->setBackColor(ed->editor->getBackColor());
  ed->tab1->show();
  tabbook->recalc();
  tabbook->handle(this, MKUINT(FXTabBar::ID_OPEN_FIRST + editors.no()-1, SEL_COMMAND), NULL);
  mdi->setDelegate(ed);
  ed->editor->setFocus();
  }

FXPLText * TextWindow::currentPLText() const {
  
  Editor * ed = currentEditor();
  if (ed->editor2 && ed->editor2->isActive())
    return ed->editor2;
  return ed->editor;
  }



// About box
long TextWindow::onCmdAbout(FXObject*,FXSelector,void*){
  FXGIFIcon icon(getApp(),bigpenguin);
  FXMessageBox about(this,"About Text Edit",
    "Text Editor Demonstrates the *other* FOX Text Widget (FXPLText et al.).\n\n"
    "Using the FOX C++ GUI Library (http://www.cfdrc.com/FOX/fox.html)\n\n"
    "Copyright (C) 2000 Jeroen van der Zijp (jvz@cfdrc.com)\n"
    "Parts Copyright (C) 2000 Stephen J. Hardy (tekno@users.sourceforge.net)\n",
    &icon,
    MBOX_OK|DECOR_TITLE|DECOR_BORDER);
  about.execute();  
  return 1;
  }


// Change font
long TextWindow::onCmdFont(FXObject*,FXSelector,void*){
  FXFixedFontDialog fontdlg(this,"Change Font",DECOR_BORDER|DECOR_TITLE);
  FXPLText * editor = currentPLText();
  FXFont *oldfont=editor->getFont();
  FXFontDesc fontdesc;
  oldfont->getFontDesc(fontdesc);
  fontdlg.setFontSelection(fontdesc);
  if(fontdlg.execute()){
    fontdlg.getFontSelection(fontdesc);
    FXFont *newfont=new FXFont(getApp(),fontdesc);
    newfont->create();
    editor->setFont(newfont, newfont);
    if(oldfont!=getApp()->getNormalFont()){
      // Don't delete the font, because it may be in use by other editors.  The memory
      // leak will be quite small, since users don't change fonts very often.
      // delete oldfont;
      }
    }
  editor->setFocus();
  return 1;
  }



// Toggle autosave mode
long TextWindow::onCmdAutosave(FXObject*,FXSelector,void*){
  autosave ^= 1;
  if (autosave)
    savetimer = getApp()->addTimeout(SAVE_MS, this, ID_SAVETIMER);
  else
    if (savetimer)
      getApp()->removeTimeout(savetimer);
  return 1;
  }


// Update toggle autosave mode
long TextWindow::onUpdAutosave(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(autosave ? ID_CHECK : ID_UNCHECK,SEL_COMMAND),NULL);
  return 1;
  }


// Toggle split mode
long TextWindow::onCmdSplit(FXObject*,FXSelector,void*){
  currentEditor()->toggleSplit();
  return 1;
  }


// Update toggle split mode
long TextWindow::onUpdSplit(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(currentEditor()->isSplit() ? ID_CHECK : ID_UNCHECK, SEL_COMMAND), NULL);
  return 1;
  }

// Toggle greek view
long TextWindow::onCmdGreek(FXObject*,FXSelector,void*){
  currentEditor()->toggleGreek();
  return 1;
  }


// Update greek view
long TextWindow::onUpdGreek(FXObject* sender,FXSelector,void*){
  sender->handle(this,MKUINT(currentEditor()->isGreek() ? ID_CHECK : ID_UNCHECK, SEL_COMMAND), NULL);
  return 1;
  }


// Revert to saved
long TextWindow::onCmdRevert(FXObject*,FXSelector,void*){
  FXFileTextSrc * textsrc = currentTextSrc();
  textsrc->revert();
  currentPLText()->setFocus();
  return 1;
  }


// Revert Update
long TextWindow::onUpdRevert(FXObject* sender,FXSelector,void* ptr){
  FXFileTextSrc * textsrc = currentTextSrc();
  if(modified != textsrc->isModified()){
    modified = textsrc->isModified();
    FXuint msg = modified ? FXWindow::ID_ENABLE : FXWindow::ID_DISABLE;
    sender->handle(this, MKUINT(msg, SEL_COMMAND), ptr);
    }
  return 1;
  }

long TextWindow::onCmdOpenButton(FXObject*o,FXSelector s,void*) {
  // This is a stub to ensure parameter is null
  return onCmdOpen(o, s, NULL);
  }

// Open
long TextWindow::onCmdOpen(FXObject*,FXSelector,void*ptr){ 
  FXFileTextSrc * textsrc = currentTextSrc();
  FXchar * name = (FXchar *)ptr;
  textsrc->setPatterns(patterns);
  FXint rc = textsrc->open(this, TRUE, name ? name : "");
  if (!rc || rc == ENOENT) {
    // We tolerate file not found, since the text source just sets empty contents
    // after warning the user.
    filename = textsrc->getFile();
    dirlist->setCurrentFile(filename);
    mrufiles.appendFile(filename);
    readRegistry(currentEditor(), textsrc->isShell() ? FXString("_shell") : FXFile::extension(filename));
    currentPLText()->setCursorPos(0);	// Move cursor to top (by default is at bottom)
    currentPLText()->makePositionVisible(0);
    }
  currentPLText()->setFocus();
  return 1;
  }


// Open recent file
long TextWindow::onCmdRecentFile(FXObject*,FXSelector,void* ptr){
  handle(this, MKUINT(ID_OPEN, SEL_COMMAND), ptr);
  return 1;
  }


// Save
long TextWindow::onCmdSave(FXObject* sender,FXSelector sel,void* ptr){
  FXFileTextSrc * textsrc = currentTextSrc();
  textsrc->save(this, TRUE);
  currentPLText()->setFocus();
  return 1;
  }


// Save Update
long TextWindow::onUpdSave(FXObject* sender,FXSelector,void* ptr){
  FXFileTextSrc * textsrc = currentTextSrc();
  FXuint msg = textsrc->isModified() && textsrc->isModifiable() ? FXWindow::ID_ENABLE : FXWindow::ID_DISABLE;
  sender->handle(this,MKUINT(msg,SEL_COMMAND),ptr);
  return 1;
  }


// Save As
long TextWindow::onCmdSaveAs(FXObject*,FXSelector,void*){
  FXFileTextSrc * textsrc = currentTextSrc();
  if (!textsrc->saveAs(this, TRUE, FXString(""))) {
    dirlist->setCurrentFile(textsrc->getFile());
    mrufiles.appendFile(textsrc->getFile());
    }
  currentPLText()->setFocus();
  return 1;
  }


// New
long TextWindow::onCmdNew(FXObject*,FXSelector,void*){
  newEditor(FXString(""));
  return 1;
  }

// New
long TextWindow::onCmdNewShell(FXObject*,FXSelector,void*){
  newEditor(FXString("shell://"));  // Pseudo-URL indicates connect to default shell
  return 1;
  }


// Quit
long TextWindow::onCmdQuit(FXObject*,FXSelector,void*){
  FXint i;
  
  for (i = 0; i < editors.no(); i++) {
    Editor * ed = editors[i];
    FXFileTextSrc * textsrc = ed->textsrc;
    if (textsrc->isModified()) {
      // Put each queried file on top of tabbook stack so user can see what is being written
      tabbook->handle(this, MKUINT(FXTabBar::ID_OPEN_FIRST+i, SEL_COMMAND), NULL);
      if (textsrc->maybeSave(this, TRUE)) return 1;
      }
    }
  
  // Write settings automatically (or not, as the case may be...)
  //writeRegistry();

  // Quit
  getApp()->exit(0);
  return 1;
  }



// Autosave timer
long TextWindow::onTimer(FXObject* sender,FXSelector,void*){
  FXint err;
  FXint inError = 0;
  for (FXint i = 0; i < editors.no(); i++) {
    Editor * ed = editors[i];
    FXFileTextSrc * textsrc = ed->textsrc;
    if (textsrc->isModified(2)) {
      if ((err = textsrc->autosave())) {
        autosave = FALSE;
        inError++;
        textsrc->showError(this, err, "Error in autosave", textsrc->getAutosaveFile().text());
        }
      }
    }
  if (!inError)
    savetimer = getApp()->addTimeout(SAVE_MS, this, ID_SAVETIMER);
  return 1;
  }


// Update title
long TextWindow::onUpdTitle(FXObject* sender,FXSelector,void*){
  FXFileTextSrc * textsrc = currentTextSrc();
  FXString title = textsrc->getFile();
  FXString tabtitle = FXFile::name(title);
  FXTabItem * tab1 = currentTab();
  if (textsrc->isModified()) { title += "*"; tabtitle += "*"; }
  tab1->setText(tabtitle);
#ifdef WIN32
  if (textsrc->getMode() == FXFileTextSrc::LF) title += " (Unix)";
#else
  if (textsrc->getMode() == FXFileTextSrc::CRLF) title += " (DOS)";
#endif
  else if (textsrc->getMode() == FXFileTextSrc::BINARY) title += " (BIN)";
  sender->handle(this,MKUINT(FXWindow::ID_SETSTRINGVALUE,SEL_COMMAND),(void*)&title);
  return 1;
  }


// Print the text
long TextWindow::onCmdPrint(FXObject*,FXSelector,void*){
  FXPrintDialog dlg(this,"Print File");
  FXPrinter printer;
  if(dlg.execute()){
    dlg.getPrinter(printer);
    FXTRACE((100,"Printer = %s\n",printer.name.text()));
    }
  return 1;
  }


// Command from the tree list
long TextWindow::onCmdTreeList(FXObject*,FXSelector,void* ptr){
  FXTreeItem *item=(FXTreeItem*)ptr;

  if(!item || !dirlist->isItemFile(item)) return 1;
  filename=dirlist->getItemPathname(item);
  mrufiles.appendFile(filename);
  handle(this, MKUINT(ID_OPEN,SEL_COMMAND), (void *)filename.text()); 
  return 1;
  }


// Change both text and tree background color
long TextWindow::onCmdBackColor(FXObject*,FXSelector,void* ptr){
  //FXPLText * editor = currentPLText();
  //editor->setBackColor((FXColor)(long)ptr);
  if (!(FXFile::extension(currentTextSrc()->getFile()).text()[0]))
    dirlist->setBackColor((FXColor)(long)ptr);
  currentEditor()->setBackColor((FXColor)(long)ptr); 
  return 1;
  }


// Update background color
long TextWindow::onUpdBackColor(FXObject* sender,FXSelector,void* ptr){
  mdi->handle(sender,MKUINT(FXPLText::ID_BACK_COLOR,SEL_UPDATE),ptr);
  return 1;
  }

// Change both text and tree selected background color
long TextWindow::onCmdSelBackColor(FXObject*,FXSelector,void* ptr){
  FXPLText * editor = currentPLText();
  editor->setSelBackColor((FXColor)(long)ptr);
  if (!(FXFile::extension(currentTextSrc()->getFile()).text()[0]))
    dirlist->setSelBackColor((FXColor)(long)ptr);
  return 1;
  }


// Update selected background color
long TextWindow::onUpdSelBackColor(FXObject* sender,FXSelector,void* ptr){
  mdi->handle(sender,MKUINT(FXPLText::ID_SELBACK_COLOR,SEL_UPDATE),ptr);
  return 1;
  }


// Change both text and tree text color
long TextWindow::onCmdTextColor(FXObject*,FXSelector,void* ptr){
  FXPLText * editor = currentPLText();
  editor->setTextColor((FXColor)(long)ptr);
  if (!(FXFile::extension(currentTextSrc()->getFile()).text()[0]))
    dirlist->setTextColor((FXColor)(long)ptr);
  return 1;
  }


// Forward GUI update to text widget
long TextWindow::onUpdTextColor(FXObject* sender,FXSelector,void* ptr){
  mdi->handle(sender,MKUINT(FXPLText::ID_TEXT_COLOR,SEL_UPDATE),ptr);
  return 1;
  }


// Change both text and tree
long TextWindow::onCmdSelTextColor(FXObject*,FXSelector,void* ptr){
  FXPLText * editor = currentPLText();
  editor->setSelTextColor((FXColor)(long)ptr);
  if (!(FXFile::extension(currentTextSrc()->getFile()).text()[0]))
    dirlist->setSelTextColor((FXColor)(long)ptr);
  return 1;
  }


// Forward GUI update to text widget
long TextWindow::onUpdSelTextColor(FXObject* sender,FXSelector,void* ptr){
  mdi->handle(sender,MKUINT(FXPLText::ID_SELTEXT_COLOR,SEL_UPDATE),ptr);
  return 1;
  }

// Change cursor colour
long TextWindow::onCmdCursorColor(FXObject*,FXSelector,void* ptr){
  FXPLText * editor = currentPLText();
  editor->setCursorColor((FXColor)(long)ptr);
  return 1;
  }


// Forward GUI update to text widget
long TextWindow::onUpdCursorColor(FXObject* sender,FXSelector,void* ptr){
  mdi->handle(sender,MKUINT(FXPLText::ID_CURSOR_COLOR,SEL_UPDATE),ptr);
  return 1;
  }


// Start the ball rolling
void TextWindow::start(int& argc, char** argv) {
  while (argc>1) {
    handle(this, MKUINT(ID_OPEN, SEL_COMMAND), argv[1]);
    currentTab()->setText(FXFile::name(currentTextSrc()->getFile()));
    argc--;
    argv++;
    if (argc>1)
      handle(this, MKUINT(ID_NEW, SEL_COMMAND), NULL);
    }
#if (0)
  // Dummy brackets for testing
  FXFileTextSrc * textsrc = currentTextSrc();
  (new FXTextBracket(textsrc, FXPFATTR_ERRORS, NULL, 10, 15))->setVisible();
  (new FXTextBracket(textsrc, FXPFATTR_SECONDARY, NULL, 20, 40))->setVisible(); 
  (new FXTextBracket(textsrc, FXPFATTR_RETENTIVE, NULL, 60, 80))->setVisible(); 
  (new FXTextBracket(textsrc, FXPFATTR_MARKER, NULL, 100, 120))->setVisible(); 
  (new FXInputBracket(textsrc, 0, FXPFATTR_INPUT, NULL, 140, 160))->setVisible(); 
  (new FXOutputBracket(textsrc, FXPFATTR_MARKER, NULL, 180, 200))->setVisible(); 
#endif
  }

void TextWindow::writeRegistry(Editor * ed, FXString ext) {
  /* Write settings specific to file extension ext (given without the dot) */
  FXPLText * editor = ed->editor;
  char fontspec[200];
  FXFontDesc fontdesc;

  FXRegistry & reg = getApp()->reg();
  if (ext.text()[0])
    ext = "SETTINGS_" + ext;
  else
    ext = "SETTINGS";
  const FXchar * set = ext.text();
  
  
  // Write changed colors back to registry
  reg.writeColorEntry(set,"textbackground",editor->getBackColor());
  reg.writeColorEntry(set,"textforeground",editor->getTextColor());
  reg.writeColorEntry(set,"textselbackground",editor->getSelBackColor());
  reg.writeColorEntry(set,"textselforeground",editor->getSelTextColor());
  reg.writeColorEntry(set,"cursorcolor",editor->getCursorColor());
    
  
  FXuint s = editor->getTextStyle();
  // Options
  reg.writeIntEntry(set,"autoindent",!!(s & PLTEXT_AUTO_INDENT));
  reg.writeIntEntry(set,"showtabs",!!(s & PLTEXT_SHOW_TABS));
  reg.writeIntEntry(set,"showaux",!!(s & PLTEXT_SHOW_AUX));
  reg.writeIntEntry(set,"tabcols",editor->getPageFormat()->getTabColumns());
  
  // Font
  editor->getFont()->getFontDesc(fontdesc);
  fxunparsefontdesc(fontspec,fontdesc);
  reg.writeStringEntry(set,"font",fontspec);

  }


void TextWindow::writeRegistry() {
  /* We write overall (default) settings to [SETTINGS], plus for each unique file extension
     we write to [SETTINGS_ext] those settings which may be dependent on file extensions i.e.
     colours and font.
  */
  FXint i, j;
  FXRegistry & reg = getApp()->reg();
  FXchar * set = "SETTINGS";
  
  // Write new window size back to registry
  reg.writeIntEntry(set,"x",getX());
  reg.writeIntEntry(set,"y",getY());
  reg.writeIntEntry(set,"width",getWidth());
  reg.writeIntEntry(set,"height",getHeight());
  
  // Were showing hidden files
  reg.writeIntEntry(set,"showhiddenfiles",dirlist->showHiddenFiles());
  
  // Was tree shown?
  reg.writeIntEntry(set,"hidetree",!treebox->shown());
  
  // Width of tree
  reg.writeIntEntry(set,"treewidth",treebox->getWidth());
  
  // Options
  reg.writeIntEntry(set,"autosave",autosave);
  
  // Write for each unique file extension
  for (i = 0; i < editors.no(); i++) {
    Editor * ed = editors[i];
    FXString ext = ed->textsrc->isShell() ? FXString("_shell") : FXFile::extension(ed->textsrc->getFile());
    for (j = 0; j < i; j++) {
      FXString ext2 = editors[j]->textsrc->isShell() ? FXString("_shell") : FXFile::extension(editors[j]->textsrc->getFile());
      if (ext == ext2)
        break;
      }
    if (j == i)
      // First entry with this extension
      writeRegistry(ed, ext);
    }
     
  }


void TextWindow::readRegistry(Editor * ed, FXString ext) {
  /* Read settings which may be dependent on extension.  Falls back to [SETTINGS] if the
     settings don't exist for this extension. */
  FXColor textback,textfore,textselback,textselfore,cursor;
  FXint autoindent,showtabs,showaux,tabcols;
  FXFontDesc fontdesc;
  FXFont *font;
  FXPLText * editor = ed->editor;
  if (ext.text()[0])
    ext = "SETTINGS_" + ext;
  else
    ext = "SETTINGS";
  const FXchar * set = ext.text();
  FXRegistry & reg = getApp()->reg();
  
  if (!reg.existingSection(set))
    set = "SETTINGS";
  
  // Get colors
  textback = reg.readColorEntry(set,"textbackground",FXRGB(255,255,255));
  textfore = reg.readColorEntry(set,"textforeground",FXRGB(0,0,0));
  textselback = reg.readColorEntry(set,"textselbackground",FXRGB(0,0,128));
  textselfore = reg.readColorEntry(set,"textselforeground",FXRGB(255,255,255));
  cursor = reg.readColorEntry(set,"cursorcolor",FXRGB(212,140,0));
  
  // Font
  if(fxparsefontdesc(fontdesc,
       reg.readStringEntry(set,"font",""))){
    font=new FXFont(getApp(),fontdesc);
    font->create();
    editor->setFont(font,font);
    }
    
  autoindent = reg.readIntEntry(set,"autoindent",0);
  showtabs = reg.readIntEntry(set,"showtabs",0);
  showaux = reg.readIntEntry(set,"showaux",0);
  tabcols = reg.readIntEntry(set,"tabcols",8);

  editor->setTextColor(textfore);
  //editor->setBackColor(textback);
  ed->setBackColor(textback);
  editor->setSelBackColor(textselback);
  editor->setSelTextColor(textselfore);
  editor->setCursorColor(cursor);

  FXuint s = editor->getTextStyle();
  s &= ~(PLTEXT_AUTO_INDENT|PLTEXT_SHOW_AUX|PLTEXT_SHOW_TABS);
  editor->setTextStyle(s | 
                       (autoindent?PLTEXT_AUTO_INDENT:0) | 
                       (showaux?PLTEXT_SHOW_AUX:0) | 
                       (showtabs?PLTEXT_SHOW_TABS:0));
  editor->getPageFormat()->setTabColumns(tabcols);
      
   
  }

// Create and show window
void TextWindow::create() {
  FXColor textback,textfore,textselback,textselfore;
  FXint ww,hh,xx,yy,treewidth,hidetree,hiddenfiles;
  Editor * ed = currentEditor();
  FXRegistry & reg = getApp()->reg();
  FXchar * set = "SETTINGS";
  
  
  // Get size
  xx = reg.readIntEntry(set,"x",600);
  yy = reg.readIntEntry(set,"y",400);
  ww = reg.readIntEntry(set,"width",600);
  hh = reg.readIntEntry(set,"height",400);

  // Hidden files shown
  hiddenfiles = reg.readIntEntry(set,"showhiddenfiles",FALSE);
  dirlist->showHiddenFiles(hiddenfiles);
  
  // Showing the tree?
  hidetree = reg.readIntEntry(set,"hidetree",TRUE);

  // Width of tree
  treewidth = reg.readIntEntry(set,"treewidth",100);

  // Other options
  autosave = reg.readIntEntry(set,"autosave",0);

  // Change the colors
  textback = reg.readColorEntry(set,"textbackground",FXRGB(255,255,255));
  textfore = reg.readColorEntry(set,"textforeground",FXRGB(0,0,0));
  textselback = reg.readColorEntry(set,"textselbackground",FXRGB(0,0,128));
  textselfore = reg.readColorEntry(set,"textselforeground",FXRGB(255,255,255));
  dirlist->setTextColor(textfore);
  dirlist->setBackColor(textback);
  dirlist->setSelBackColor(textselback);
  dirlist->setSelTextColor(textselfore);
   
  // Hide tree if asked for
  if(hidetree) treebox->hide();
  
  // Set tree width
  treebox->setWidth(treewidth);
  
  // Open toward file
  dirlist->setCurrentFile(filename);
        
  // Reposition window
  position(xx,yy,ww,hh);
  
  // Read settings for editor
  readRegistry(ed, FXString(""));
  
  FXMainWindow::create();

  if (autosave)
    savetimer = getApp()->addTimeout(SAVE_MS, this, ID_SAVETIMER);

  show();
  }


// Close editor cluster
long TextWindow::onCmdClose(FXObject*,FXSelector,void*){
  if (editors.no() <= 1)
    // Only one editor open - handle as for quit.
    return handle(this, MKUINT(ID_QUIT, SEL_COMMAND), NULL);
  if (currentTextSrc()->maybeSave(this, TRUE)) return 1;
  FXint which;
  if ((which = editors.findf(currentEditor())) < 0) return 1;
  FXint next = which == editors.no()-1 ? 0 : which;
  editors.remove(which);
  delete currentEditor();
  tabbook->recalc();
  tabbook->handle(this, MKUINT(FXTabBar::ID_OPEN_FIRST+next, SEL_COMMAND), NULL);
  mdi->setDelegate(editors[next]);
  return 1;
  }


// Cause SIGSEGV
long TextWindow::onCmdCrash(FXObject*,FXSelector,void*){
  *(char *)0 = 0;
  return 1;
  }


// From tabbook - selected item changed
long TextWindow::onCmdTab(FXObject*,FXSelector,void*ptr){
  FXint which = (FXint)ptr;
  mdi->setDelegate(editors[which]);
  editors[which]->editor->setFocus();
  return 1;
  }

// Save settings
long TextWindow::onCmdSaveSettings(FXObject*,FXSelector,void*){
  writeRegistry();
  getApp()->reg().write();
  return 1;
  }

// Set editor appendable (last line editable) and move cursor.
long TextWindow::onCmdSetAppendable(FXObject*,FXSelector,void*){
  currentPLText()->setAppendable();
  return 1;
  }

// Set editor prependable (first line editable) and move cursor.
long TextWindow::onCmdSetPrependable(FXObject*,FXSelector,void*){
  currentPLText()->setPrependable();
  return 1;
  }


// Set file save modes for current editor.
long TextWindow::onCmdMode(FXObject*,FXSelector sel,void*){
  currentTextSrc()->setMode((FXFileTextSrc::le_mode)(SELID(sel) - ID_MODE_AUTO + FXFileTextSrc::AUTO));
  return 1;
  }


long TextWindow::onUpdMode(FXObject*sender,FXSelector sel,void*){
  FXMenuCommand *cmd = (FXMenuCommand*)sender;
  (FXFileTextSrc::le_mode)(SELID(sel) - ID_MODE_AUTO + FXFileTextSrc::AUTO) == 
    currentTextSrc()->getMode() ?
      cmd->checkRadio() : 
      cmd->uncheckRadio();
  return 1;
  }




/*******************************************************************************/


// Start the whole thing
int main(int argc,char *argv[]){

  // Make application
  FXApp* application=new FXApp("PLTextEdit3","Test");
  
  // Open display
  application->init(argc,argv);

  // Make window
  TextWindow* window=new TextWindow(application);
  
  // Handle interrupt to save stuff nicely
  application->addSignal(SIGINT,window,TextWindow::ID_QUIT);

  // Create it
  application->create();
  
  // Start
  window->start(argc,argv);

  // Run
  application->run();

  return 0;
  }


